# pylint: disable=line-too-long

""" AvroToCSharp class for converting Avro schema to C# classes """

import json
import os
import re
from typing import Any, Dict, List, Tuple, Union, cast
import uuid

from avrotize.common import build_flat_type_dict, inline_avro_references, is_generic_avro_type, pascal, process_template
from avrotize.constants import (
    CSHARP_AVRO_VERSION,
    NEWTONSOFT_JSON_VERSION,
    SYSTEM_TEXT_JSON_VERSION,
    SYSTEM_MEMORY_DATA_VERSION,
    PROTOBUF_NET_VERSION,
    MSGPACK_VERSION,
    NUNIT_VERSION,
    NUNIT_ADAPTER_VERSION,
    MSTEST_SDK_VERSION,
    COVERLET_VERSION,
)
import glob

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None


INDENT = '    '

AVRO_CLASS_PREAMBLE = \
"""
public {type_name}(global::Avro.Generic.GenericRecord obj)
{
    global::Avro.Specific.ISpecificRecord self = this;
    for (int i = 0; obj.Schema.Fields.Count > i; ++i)
    {
        self.Put(i, obj.GetValue(i));
    }
}
"""

class AvroToCSharp:
    """ Converts Avro schema to C# classes """

    def __init__(self, base_namespace: str = '') -> None:
        self.base_namespace = base_namespace
        self.project_name: str = ''  # Optional explicit project name, separate from namespace
        self.schema_doc: JsonNode = None
        self.output_dir = os.getcwd()
        self.pascal_properties = False
        self.system_text_json_annotation = False
        self.newtonsoft_json_annotation = False
        self.system_xml_annotation = False
        self.msgpack_annotation = False
        self.avro_annotation = False
        self.protobuf_net_annotation = False
        self.generated_types: Dict[str,str] = {}
        self.generated_avro_types: Dict[str, Dict[str, Union[str, Dict, List]]] = {}
        self.type_dict: Dict[str, Dict] = {}

    def get_qualified_name(self, namespace: str, name: str) -> str:
        """ Concatenates namespace and name with a dot separator """
        return f"{namespace}.{name}" if namespace != '' else name

    def concat_namespace(self, namespace: str, name: str) -> str:
        """ Concatenates namespace and name with a dot separator """
        if namespace and name:
            return f"{namespace}.{name}"
        elif namespace:
            return namespace
        else:
            return name

    def map_primitive_to_csharp(self, avro_type: str) -> str:
        """ Maps Avro primitive types to C# types """
        mapping = {
            'null': 'void',  # Placeholder, actual handling for nullable types is in the union logic
            'boolean': 'bool',
            'int': 'int',
            'long': 'long',
            'float': 'float',
            'double': 'double',
            'bytes': 'byte[]',
            'string': 'string',
        }
        qualified_class_name = 'global::'+self.get_qualified_name(pascal(self.base_namespace), pascal(avro_type))
        if qualified_class_name in self.generated_avro_types:
            result = qualified_class_name
        else:
            result = mapping.get(avro_type, 'object')
        return result

    def is_csharp_reserved_word(self, word: str) -> bool:
        """ Checks if a word is a reserved C# keyword """
        reserved_words = [
            'abstract', 'as', 'base', 'bool', 'break', 'byte', 'case', 'catch', 'char', 'checked', 'class', 'const',
            'continue', 'decimal', 'default', 'delegate', 'do', 'double', 'else', 'enum', 'event', 'explicit', 'extern',
            'false', 'finally', 'fixed', 'float', 'for', 'foreach', 'goto', 'if', 'implicit', 'in', 'int', 'interface',
            'internal', 'is', 'lock', 'long', 'namespace', 'new', 'null', 'object', 'operator', 'out', 'override',
            'params', 'private', 'protected', 'public', 'readonly', 'ref', 'return', 'sbyte', 'sealed', 'short', 'sizeof',
            'stackalloc', 'static', 'string', 'struct', 'switch', 'this', 'throw', 'true', 'try', 'typeof', 'uint', 'ulong',
            'unchecked', 'unsafe', 'ushort', 'using', 'virtual', 'void', 'volatile', 'while'
        ]
        return word in reserved_words

    def is_csharp_primitive_type(self, csharp_type: str) -> bool:
        """ Checks if an Avro type is a C# primitive type """
        if csharp_type.endswith('?'):
            csharp_type = csharp_type[:-1]
        return csharp_type in ['null', 'bool', 'int', 'long', 'float', 'double', 'bytes', 'string', 'DateTime', 'decimal', 'short', 'sbyte', 'ushort', 'uint', 'ulong', 'byte[]', 'object']

    def map_csharp_primitive_to_clr_type(self, cs_type: str) -> str:
        """ Maps C# primitive types to CLR types"""
        map = {
            "int": "Int32",
            "long": "Int64",
            "float": "Single",
            "double": "Double",
            "decimal": "Decimal",
            "short": "Int16",
            "sbyte": "SByte",
            "ushort": "UInt16",
            "uint": "UInt32",
            "ulong": "UInt64"
        }
        return map.get(cs_type, cs_type)

    def convert_avro_type_to_csharp(self, class_name: str, field_name: str, avro_type: JsonNode, parent_namespace: str) -> str:
        """ Converts Avro type to C# type """
        if isinstance(avro_type, str):
            return self.map_primitive_to_csharp(avro_type)
        elif isinstance(avro_type, list):
            # Handle nullable types and unions
            if is_generic_avro_type(avro_type):
                return 'Dictionary<string, object>'
            else:
                non_null_types = [t for t in avro_type if t != 'null']
                if len(non_null_types) == 1:
                    # Nullable type
                    return f"{self.convert_avro_type_to_csharp(class_name, field_name, non_null_types[0], parent_namespace)}?"
                else:
                    return self.generate_embedded_union(class_name, field_name, non_null_types, parent_namespace, write_file=True)
        elif isinstance(avro_type, dict):
            # Handle complex types: records, enums, arrays, and maps
            if avro_type['type'] in ['record', 'enum']:
                return self.generate_class_or_enum(avro_type, parent_namespace, write_file=True)
            elif avro_type['type'] == 'array':
                return f"List<{self.convert_avro_type_to_csharp(class_name, field_name+'List', avro_type['items'], parent_namespace)}>"
            elif avro_type['type'] == 'map':
                return f"Dictionary<string, {self.convert_avro_type_to_csharp(class_name, field_name, avro_type['values'], parent_namespace)}>"
            return self.convert_avro_type_to_csharp(class_name, field_name, avro_type['type'], parent_namespace)
        return 'object'

    def generate_class_or_enum(self, avro_schema: Dict, parent_namespace: str, write_file: bool = True) -> str:
        """ Generates a Class or Enum """
        if avro_schema['type'] == 'record':
            return self.generate_class(avro_schema, parent_namespace, write_file)
        elif avro_schema['type'] == 'enum':
            return self.generate_enum(avro_schema, parent_namespace, write_file)
        return ''

    def generate_class(self, avro_schema: Dict, parent_namespace: str, write_file: bool) -> str:
        """ Generates a Class """
        class_definition = ''
        avro_namespace = avro_schema.get('namespace', parent_namespace)
        if not 'namespace' in avro_schema:
            avro_schema['namespace'] = parent_namespace
        xml_namespace = avro_schema.get('xmlns', None)
        namespace = pascal(self.concat_namespace(self.base_namespace, avro_namespace))
        class_name = pascal(avro_schema['name'])
        ref = 'global::'+self.get_qualified_name(namespace, class_name)
        if ref in self.generated_types:
            return ref

        class_definition += f"/// <summary>\n/// { avro_schema.get('doc', class_name ) }\n/// </summary>\n"

        # Add XML serialization attribute for the class if enabled
        if self.system_xml_annotation:
            if xml_namespace:
                class_definition += f"[XmlRoot(\"{class_name}\", Namespace=\"{xml_namespace}\")]\n"
            else:
                class_definition += f"[XmlRoot(\"{class_name}\")]\n"

        if self.protobuf_net_annotation:
            class_definition += "[ProtoContract]\n"

        # Add MessagePack serialization attribute for the class if enabled
        if self.msgpack_annotation:
            class_definition += "[MessagePackObject]\n"

        fields_str = [self.generate_property(index + 1, field, class_name, avro_namespace) for index, field in enumerate(avro_schema.get('fields', []))]
        class_body = "\n".join(fields_str)
        class_definition += f"public partial class {class_name}"
        if self.avro_annotation:
            class_definition += " : global::Avro.Specific.ISpecificRecord"
        class_definition += "\n{\n"+class_body
        class_definition += f"\n{INDENT}/// <summary>\n{INDENT}/// Default constructor\n{INDENT}///</summary>\n"
        class_definition += f"{INDENT}public {class_name}()\n{INDENT}{{\n{INDENT}}}"
        if self.avro_annotation:
            class_definition += f"\n\n{INDENT}/// <summary>\n{INDENT}/// Constructor from Avro GenericRecord\n{INDENT}///</summary>\n"
            class_definition += f"{INDENT}public {class_name}(global::Avro.Generic.GenericRecord obj)\n{INDENT}{{\n"
            class_definition += f"{INDENT*2}global::Avro.Specific.ISpecificRecord self = this;\n"
            class_definition += f"{INDENT*2}for (int i = 0; obj.Schema.Fields.Count > i; ++i)\n{INDENT*2}{{\n"
            class_definition += f"{INDENT*3}self.Put(i, obj.GetValue(i));\n{INDENT*2}}}\n{INDENT}}}\n"
        if self.avro_annotation:
            
            local_avro_schema = inline_avro_references(avro_schema.copy(), self.type_dict, '')
            avro_schema_json = json.dumps(local_avro_schema)
            # wrap schema at 80 characters
            avro_schema_json = avro_schema_json.replace('"', '§')
            avro_schema_json = f"\"+\n{INDENT}\"".join(
                [avro_schema_json[i:i+80] for i in range(0, len(avro_schema_json), 80)])
            avro_schema_json = avro_schema_json.replace('§', '\\"')
            class_definition += f"\n\n{INDENT}/// <summary>\n{INDENT}/// Avro schema for this class\n{INDENT}/// </summary>"
            class_definition += f"\n{INDENT}public static global::Avro.Schema AvroSchema = global::Avro.Schema.Parse(\n{INDENT}\"{avro_schema_json}\");\n"
            class_definition += f"\n{INDENT}global::Avro.Schema global::Avro.Specific.ISpecificRecord.Schema => AvroSchema;\n"
            get_method = f"{INDENT}object global::Avro.Specific.ISpecificRecord.Get(int fieldPos)\n" + \
                INDENT+"{"+f"\n{INDENT*2}switch (fieldPos)\n{INDENT*2}" + "{"
            put_method = f"{INDENT}void global::Avro.Specific.ISpecificRecord.Put(int fieldPos, object fieldValue)\n" + \
                INDENT+"{"+f"\n{INDENT*2}switch (fieldPos)\n{INDENT*2}"+"{"
            for pos, field in enumerate(avro_schema.get('fields', [])):
                field_name = field['name']
                if self.is_csharp_reserved_word(field_name):
                    field_name = f"@{field_name}"
                field_type = self.convert_avro_type_to_csharp(class_name, field_name, field['type'], avro_namespace)
                if self.pascal_properties:
                    field_name = pascal(field_name)
                if field_name == class_name:
                    field_name += "_"
                if field_type in self.generated_types:
                    if self.generated_types[field_type] == "union":
                        get_method += f"\n{INDENT*3}case {pos}: return this.{field_name}?.ToObject();"
                        put_method += f"\n{INDENT*3}case {pos}: this.{field_name} = {field_type}.FromObject(fieldValue); break;"
                    elif self.generated_types[field_type] == "enum":
                        get_method += f"\n{INDENT*3}case {pos}: return ({field_type})this.{field_name};"
                        put_method += f"\n{INDENT*3}case {pos}: this.{field_name} = fieldValue is global::Avro.Generic.GenericEnum?Enum.Parse<{field_type}>(((global::Avro.Generic.GenericEnum)fieldValue).Value):({field_type})fieldValue; break;"
                    elif self.generated_types[field_type] == "class":
                        get_method += f"\n{INDENT*3}case {pos}: return this.{field_name};"
                        put_method += f"\n{INDENT*3}case {pos}: this.{field_name} = fieldValue is global::Avro.Generic.GenericRecord?new {field_type}((global::Avro.Generic.GenericRecord)fieldValue):({field_type})fieldValue; break;"
                else:
                    get_method += f"\n{INDENT*3}case {pos}: return this.{field_name};"
                    if field_type.startswith("List<"):
                        inner_type = field_type.strip()[5:-2] if field_type[-1] == '?' else field_type[5:-1]
                        if inner_type in self.generated_types:
                            if self.generated_types[inner_type] == "class":
                                put_method += f"\n{INDENT*3}case {pos}: this.{field_name} = fieldValue is Object[]?((Object[])fieldValue).Select(x => new {inner_type}((global::Avro.Generic.GenericRecord)x)).ToList():({field_type})fieldValue; break;"
                            else:
                                put_method += f"\n{INDENT*3}case {pos}: this.{field_name} = fieldValue is Object[]?((Object[])fieldValue).Select(x => ({inner_type})x).ToList():({field_type})fieldValue; break;"
                        else:
                            put_method += f"\n{INDENT*3}case {pos}: this.{field_name} = fieldValue is Object[]?((Object[])fieldValue).Select(x => ({inner_type})x).ToList():({field_type})fieldValue; break;"
                    else:
                        put_method += f"\n{INDENT*3}case {pos}: this.{field_name} = ({field_type})fieldValue; break;"
            get_method += f"\n{INDENT*3}default: throw new global::Avro.AvroRuntimeException($\"Bad index {{fieldPos}} in Get()\");"
            put_method += f"\n{INDENT*3}default: throw new global::Avro.AvroRuntimeException($\"Bad index {{fieldPos}} in Put()\");"
            get_method += "\n"+INDENT+INDENT+"}\n"+INDENT+"}"
            put_method += "\n"+INDENT+INDENT+"}\n"+INDENT+"}"
            class_definition += f"\n{get_method}\n{put_method}\n"

        # emit helper methods
        class_definition += process_template(
            "avrotocsharp/dataclass_core.jinja",
            class_name=class_name,
            avro_annotation=self.avro_annotation,
            protobuf_net_annotation=self.protobuf_net_annotation,
            system_text_json_annotation=self.system_text_json_annotation,
            newtonsoft_json_annotation=self.newtonsoft_json_annotation,
            system_xml_annotation=self.system_xml_annotation,
            msgpack_annotation=self.msgpack_annotation,
            json_match_clauses=self.create_is_json_match_clauses(avro_schema, avro_namespace, class_name)
        )

        # emit Equals and GetHashCode for value equality
        class_definition += self.generate_equals_and_gethashcode(avro_schema, class_name, avro_namespace)

        class_definition += "\n"+"}"

        if write_file:
            self.write_to_file(namespace, class_name, class_definition)

        self.generated_types[ref] = "class"
        self.generated_avro_types[ref] = avro_schema
        return ref

    def create_is_json_match_clauses(self, avro_schema, parent_namespace, class_name) -> List[str]:
        """ Generates the IsJsonMatch method for System.Text.Json """
        clauses: List[str] = []
        for field in avro_schema.get('fields', []):
            field_name = field['name']
            if self.is_csharp_reserved_word(field_name):
                field_name = f"@{field_name}"
            if field_name == class_name:
                field_name += "_"
            field_type = self.convert_avro_type_to_csharp(
                    class_name, field_name, field['type'], parent_namespace)
            clauses.append(self.get_is_json_match_clause(class_name, field_name, field_type))
        if len(clauses) == 0:
            clauses.append("true")
        return clauses

    def get_is_json_match_clause(self, class_name, field_name, field_type) -> str:
        """ Generates the IsJsonMatch clause for a field """
        class_definition = ''
        field_name_js = field_name[1:] if field_name[0] == '@' else field_name
        is_optional = field_type[-1] == '?'
        field_type = field_type[:-1] if is_optional else field_type
        if field_type == 'byte[]':
            class_definition += f"({'!' if is_optional else ''}element.TryGetProperty(\"{field_name_js}\", out System.Text.Json.JsonElement {field_name}) {'||' if is_optional else '&&'} ({field_name}.ValueKind == System.Text.Json.JsonValueKind.String){f' || {field_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        elif field_type == 'string':
            class_definition += f"({'!' if is_optional else ''}element.TryGetProperty(\"{field_name_js}\", out System.Text.Json.JsonElement {field_name}) {'||' if is_optional else '&&'} ({field_name}.ValueKind == System.Text.Json.JsonValueKind.String){f' || {field_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        elif field_type in ['int', 'long', 'float', 'double', 'decimal', 'short', 'sbyte', 'ushort', 'uint', 'ulong']:
            class_definition += f"({'!' if is_optional else ''}element.TryGetProperty(\"{field_name_js}\", out System.Text.Json.JsonElement {field_name}) {'||' if is_optional else '&&'} ({field_name}.ValueKind == System.Text.Json.JsonValueKind.Number){f' || {field_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        elif field_type == 'bool':
            class_definition += f"({'!' if is_optional else ''}element.TryGetProperty(\"{field_name_js}\", out System.Text.Json.JsonElement {field_name}) {'||' if is_optional else '&&'} ({field_name}.ValueKind == System.Text.Json.JsonValueKind.True || {field_name}.ValueKind == System.Text.Json.JsonValueKind.False){f' || {field_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        elif field_type.startswith("global::"):
            type_kind = self.generated_types[field_type] if field_type in self.generated_types else "class"
            if type_kind == "class":
                class_definition += f"({'!' if is_optional else ''}element.TryGetProperty(\"{field_name_js}\", out System.Text.Json.JsonElement {field_name}) {'||' if is_optional else '&&'} ({f'{field_name}.ValueKind == System.Text.Json.JsonValueKind.Null || ' if is_optional else ''}{field_type}.IsJsonMatch({field_name})))"
            elif type_kind == "enum":
                class_definition += f"({'!' if is_optional else ''}element.TryGetProperty(\"{field_name_js}\", out System.Text.Json.JsonElement {field_name}) {'||' if is_optional else '&&'} ({f'{field_name}.ValueKind == System.Text.Json.JsonValueKind.Null ||' if is_optional else ''}({field_name}.ValueKind == System.Text.Json.JsonValueKind.String && Enum.TryParse<{field_type}>({field_name}.GetString(), true, out _ ))))"
        else:
            is_union = False
            field_union = pascal(field_name)+'Union'
            if field_type == field_union:
                field_union = class_name+"."+pascal(field_name)+'Union'
                type_kind = self.generated_types[field_union] if field_union in self.generated_types else "class"
                if type_kind == "union":
                    is_union = True
                    class_definition += f"({'!' if is_optional else ''}element.TryGetProperty(\"{field_name_js}\", out System.Text.Json.JsonElement {field_name}) {'||' if is_optional else '&&'} ({f'{field_name}.ValueKind == System.Text.Json.JsonValueKind.Null || ' if is_optional else ''}{field_type}.IsJsonMatch({field_name})))"
            if not is_union:
                class_definition += f"({'!' if is_optional else ''}element.TryGetProperty(\"{field_name_js}\", out System.Text.Json.JsonElement {field_name}) {'||' if is_optional else '&&'} true )"
        return class_definition

    def get_is_json_match_clause_type(self, element_name, class_name, field_type) -> str:
        """ Generates the IsJsonMatch clause for a field """
        class_definition = ''
        is_optional = field_type[-1] == '?'
        field_type = field_type[:-1] if is_optional else field_type
        if field_type == 'byte[]':
            class_definition += f"({element_name}.ValueKind == System.Text.Json.JsonValueKind.String{f' || {element_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        elif field_type == 'string':
            class_definition += f"({element_name}.ValueKind == System.Text.Json.JsonValueKind.String{f' || {element_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        elif field_type in ['int', 'long', 'float', 'double', 'decimal', 'short', 'sbyte', 'ushort', 'uint', 'ulong']:
            class_definition += f"({element_name}.ValueKind == System.Text.Json.JsonValueKind.Number{f' || {element_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        elif field_type == 'bool':
            class_definition += f"({element_name}.ValueKind == System.Text.Json.JsonValueKind.True || {element_name}.ValueKind == System.Text.Json.JsonValueKind.False{f' || {element_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        elif field_type.startswith("global::"):
            type_kind = self.generated_types[field_type] if field_type in self.generated_types else "class"
            if type_kind == "class":
                class_definition += f"({f'{element_name}.ValueKind == System.Text.Json.JsonValueKind.Null || ' if is_optional else ''}{field_type}.IsJsonMatch({element_name}))"
            elif type_kind == "enum":
                class_definition += f"({f'{element_name}.ValueKind == System.Text.Json.JsonValueKind.Null ||' if is_optional else ''}({element_name}.ValueKind == System.Text.Json.JsonValueKind.String && Enum.TryParse<{field_type}>({element_name}.GetString(), true, out _ )))"
        else:
            is_union = False
            field_union = pascal(element_name)+'Union'
            if field_type == field_union:
                field_union = class_name+"."+pascal(element_name)+'Union'
                type_kind = self.generated_types[field_union] if field_union in self.generated_types else "class"
                if type_kind == "union":
                    is_union = True
                    class_definition += f"({f'{element_name}.ValueKind == System.Text.Json.JsonValueKind.Null || ' if is_optional else ''}{field_type}.IsJsonMatch({element_name})))"
            if not is_union:
                class_definition += f"({element_name}.ValueKind == System.Text.Json.JsonValueKind.Object{f' || {element_name}.ValueKind == System.Text.Json.JsonValueKind.Null' if is_optional else ''})"
        return class_definition

    def generate_equals_and_gethashcode(self, avro_schema: Dict, class_name: str, parent_namespace: str) -> str:
        """ Generates Equals and GetHashCode methods for value equality """
        code = "\n"
        fields = avro_schema.get('fields', [])
        
        if not fields:
            # Empty class - simple implementation
            code += f"{INDENT}/// <summary>\n{INDENT}/// Determines whether the specified object is equal to the current object.\n{INDENT}/// </summary>\n"
            code += f"{INDENT}public override bool Equals(object? obj)\n{INDENT}{{\n"
            code += f"{INDENT*2}return obj is {class_name};\n"
            code += f"{INDENT}}}\n\n"
            code += f"{INDENT}/// <summary>\n{INDENT}/// Serves as the default hash function.\n{INDENT}/// </summary>\n"
            code += f"{INDENT}public override int GetHashCode()\n{INDENT}{{\n"
            code += f"{INDENT*2}return 0;\n"
            code += f"{INDENT}}}\n"
            return code
        
        # Generate Equals method
        code += f"{INDENT}/// <summary>\n{INDENT}/// Determines whether the specified object is equal to the current object.\n{INDENT}/// </summary>\n"
        code += f"{INDENT}public override bool Equals(object? obj)\n{INDENT}{{\n"
        code += f"{INDENT*2}if (obj is not {class_name} other) return false;\n"
        
        # Build equality comparisons for each field
        equality_checks = []
        for field in fields:
            field_name = field['name']
            if self.is_csharp_reserved_word(field_name):
                field_name = f"@{field_name}"
            if self.pascal_properties:
                field_name = pascal(field_name)
            if field_name == class_name:
                field_name += "_"
            
            field_type = self.convert_avro_type_to_csharp(class_name, field_name, field['type'], parent_namespace)
            
            # Handle different types of comparisons
            if field_type == 'byte[]' or field_type == 'byte[]?':
                # Byte arrays need special handling
                equality_checks.append(f"System.Linq.Enumerable.SequenceEqual({field_name} ?? Array.Empty<byte>(), other.{field_name} ?? Array.Empty<byte>())")
            elif field_type.startswith('List<') or field_type.startswith('Dictionary<'):
                # Collections need sequence comparison
                if field_type.endswith('?'):
                    equality_checks.append(f"(({field_name} == null && other.{field_name} == null) || ({field_name} != null && other.{field_name} != null && {field_name}.SequenceEqual(other.{field_name})))")
                else:
                    equality_checks.append(f"{field_name}.SequenceEqual(other.{field_name})")
            else:
                # Use Equals for reference types, == for value types
                if field_type.endswith('?') or not self.is_csharp_primitive_type(field_type):
                    equality_checks.append(f"Equals({field_name}, other.{field_name})")
                else:
                    equality_checks.append(f"{field_name} == other.{field_name}")
        
        # Join all checks with &&
        if len(equality_checks) == 1:
            code += f"{INDENT*2}return {equality_checks[0]};\n"
        else:
            code += f"{INDENT*2}return " + f"\n{INDENT*3}&& ".join(equality_checks) + ";\n"
        
        code += f"{INDENT}}}\n\n"
        
        # Generate GetHashCode method
        code += f"{INDENT}/// <summary>\n{INDENT}/// Serves as the default hash function.\n{INDENT}/// </summary>\n"
        code += f"{INDENT}public override int GetHashCode()\n{INDENT}{{\n"
        
        # Collect field names for HashCode.Combine
        hash_fields = []
        for field in fields:
            field_name = field['name']
            if self.is_csharp_reserved_word(field_name):
                field_name = f"@{field_name}"
            if self.pascal_properties:
                field_name = pascal(field_name)
            if field_name == class_name:
                field_name += "_"
            
            field_type = self.convert_avro_type_to_csharp(class_name, field_name, field['type'], parent_namespace)
            
            # Handle special types that need custom hash code computation
            if field_type == 'byte[]' or field_type == 'byte[]?':
                hash_fields.append(f"({field_name} != null ? System.Convert.ToBase64String({field_name}).GetHashCode() : 0)")
            elif field_type.startswith('List<') or field_type.startswith('Dictionary<'):
                # For collections, compute hash from elements
                if field_type.endswith('?'):
                    hash_fields.append(f"({field_name} != null ? {field_name}.Aggregate(0, (acc, item) => HashCode.Combine(acc, item)) : 0)")
                else:
                    hash_fields.append(f"{field_name}.Aggregate(0, (acc, item) => HashCode.Combine(acc, item))")
            else:
                hash_fields.append(field_name)
        
        # HashCode.Combine supports up to 8 parameters
        if len(hash_fields) <= 8:
            code += f"{INDENT*2}return HashCode.Combine({', '.join(hash_fields)});\n"
        else:
            # For more than 8 fields, use HashCode.Add
            code += f"{INDENT*2}var hash = new HashCode();\n"
            for field in hash_fields:
                code += f"{INDENT*2}hash.Add({field});\n"
            code += f"{INDENT*2}return hash.ToHashCode();\n"
        
        code += f"{INDENT}}}\n"
        
        return code

    def generate_enum(self, avro_schema: Dict, parent_namespace: str, write_file: bool) -> str:
        """ Generates an Enum """
        enum_definition = ''
        namespace = pascal(self.concat_namespace(
            self.base_namespace, avro_schema.get('namespace', parent_namespace)))
        xml_namespace = avro_schema.get('xmlns', None)
        enum_name = pascal(avro_schema['name'])
        ref = 'global::'+self.get_qualified_name(namespace, enum_name)
        if ref in self.generated_types:
            return ref

        enum_definition += f"/// <summary>\n/// {avro_schema.get('doc', enum_name )}\n/// </summary>\n"

        # Add XML serialization attribute for the enum if enabled
        if self.system_xml_annotation:
            if xml_namespace:
                enum_definition += f"[XmlType(\"{enum_name}\", Namespace=\"{xml_namespace}\")]\n"
            else:
                enum_definition += f"[XmlType(\"{enum_name}\")]\n"

        if self.system_xml_annotation:
            symbols_str = [f"{INDENT}/// <summary>\n{INDENT}/// {symbol}\n{INDENT}/// </summary>\n{INDENT}[XmlEnum(Name=\"{symbol}\")]\n{INDENT}{symbol}" for symbol in avro_schema['symbols']]
        else:
            symbols_str = [f"{INDENT}/// <summary>\n{INDENT}/// {symbol}\n{INDENT}/// </summary>\n{INDENT}{symbol}" for symbol in avro_schema['symbols']]
        enum_body = ",\n".join(symbols_str)
        enum_definition += f"public enum {enum_name}\n{{\n{enum_body}\n}}"

        if write_file:
            self.write_to_file(namespace, enum_name, enum_definition)
        ref = 'global::'+self.get_qualified_name(namespace, enum_name)
        self.generated_types[ref] = "enum"
        self.generated_avro_types[ref] = avro_schema
        return ref

    def generate_embedded_union(self, class_name: str, field_name: str, avro_type: List, parent_namespace: str, write_file: bool) -> str:
        """ Generates an embedded Union Class """

        class_definition_ctors = class_definition_decls = class_definition_read = ''
        class_definition_write = class_definition = class_definition_toobject = ''
        class_definition_objctr = class_definition_genericrecordctor = ''
        namespace = pascal(self.concat_namespace(self.base_namespace, parent_namespace))
        list_is_json_match: List [str] = []
        union_class_name = pascal(field_name)+'Union'
        ref = class_name+'.'+union_class_name

        union_types = [self.convert_avro_type_to_csharp(class_name, field_name+"Option"+str(i), t, parent_namespace) for i,t in enumerate(avro_type)]
        for i, union_type in enumerate(union_types):
            is_dict = is_list = False
            if union_type.startswith("Dictionary<"):
                # get the type information from the dictionary
                is_dict = True
                match = re.findall(r"Dictionary<(.+)\s*,\s*(.+)>", union_type)
                union_type_name = "Map" + pascal(match[0][1].rsplit('.', 1)[-1])
            elif union_type.startswith("List<"):
                # get the type information from the list
                is_list = True
                match = re.findall(r"List<(.+)>", union_type)
                union_type_name = "Array" + pascal(match[0].rsplit('.', 1)[-1])
            elif union_type == "byte[]":
                union_type_name = "bytes"
            else:
                union_type_name = union_type.rsplit('.', 1)[-1]
            if self.is_csharp_reserved_word(union_type_name):
                union_type_name = f"@{union_type_name}"
            proto_member_name = union_type_name[1:] if union_type_name.startswith("@") else union_type_name
            class_definition_objctr += f"{INDENT*3}if (obj is {union_type})\n{INDENT*3}{{\n{INDENT*4}self.{union_type_name} = ({union_type})obj;\n{INDENT*4}return self;\n{INDENT*3}}}\n"
            if union_type in self.generated_types and self.generated_types[union_type] == "class":
                class_definition_genericrecordctor += f"{INDENT*3}if (obj.Schema.Fullname == {union_type}.AvroSchema.Fullname)\n{INDENT*3}{{\n{INDENT*4}this.{union_type_name} = new {union_type}(obj);\n{INDENT*4}return;\n{INDENT*3}}}\n"
            class_definition_ctors += \
                f"{INDENT*2}/// <summary>\n{INDENT*2}/// Constructor for {union_type_name} values\n{INDENT*2}/// </summary>\n" + \
                f"{INDENT*2}public {union_class_name}({union_type}? {union_type_name})\n{INDENT*2}{{\n{INDENT*3}this.{union_type_name} = {union_type_name};\n{INDENT*2}}}\n"
            # Add Key attribute for MessagePack serialization if enabled
            msgpack_key_attr = f"{INDENT*2}[Key({i})]\n" if self.msgpack_annotation else ""
            class_definition_decls += \
                f"{INDENT*2}/// <summary>\n{INDENT*2}/// Gets the {union_type_name} value\n{INDENT*2}/// </summary>\n" + \
                (f"{INDENT*2}[ProtoMember({i+1}, Name=\"{proto_member_name}\")]\n" if self.protobuf_net_annotation else "") + \
                msgpack_key_attr + \
                f"{INDENT*2}public {union_type}? {union_type_name} {{ get; set; }} = null;\n"
            class_definition_toobject += f"{INDENT*3}if ({union_type_name} != null) {{\n{INDENT*4}return {union_type_name};\n{INDENT*3}}}\n"

            if self.system_text_json_annotation:
                if is_dict:
                    class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.Object)\n{INDENT*3}{{\n" + \
                            f"{INDENT*4}var map = System.Text.Json.JsonSerializer.Deserialize<{union_type}>(element, options);\n" + \
                            f"{INDENT*4}if (map != null) {{ return new {union_class_name}(map); }} else {{ throw new NotSupportedException(); }};\n" + \
                            f"{INDENT*3}}}\n"
                elif is_list:
                    class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.Array)\n{INDENT*3}{{\n" + \
                            f"{INDENT*4}var map = System.Text.Json.JsonSerializer.Deserialize<{union_type}>(element, options);\n" + \
                            f"{INDENT*4}if (map != null) {{ return new {union_class_name}(map); }} else {{ throw new NotSupportedException(); }};\n" + \
                            f"{INDENT*3}}}\n"
                elif self.is_csharp_primitive_type(union_type):
                    if union_type == "byte[]":
                        class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.String)\n{INDENT*3}{{\n{INDENT*4}return new {union_class_name}(element.GetBytesFromBase64());\n{INDENT*3}}}\n"
                    if union_type == "string":
                        class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.String)\n{INDENT*3}{{\n{INDENT*4}return new {union_class_name}(element.GetString());\n{INDENT*3}}}\n"
                    elif union_type in ['int', 'long', 'float', 'double', 'decimal', 'short', 'sbyte', 'ushort', 'uint', 'ulong']:
                        class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.Number)\n{INDENT*3}{{\n{INDENT*4}return new {union_class_name}(element.Get{self.map_csharp_primitive_to_clr_type(union_type)}());\n{INDENT*3}}}\n"
                    elif union_type == "bool":
                        class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.True || element.ValueKind == System.Text.Json.JsonValueKind.False)\n{INDENT*2}{{\n{INDENT*3}return new {union_class_name}(element.GetBoolean());\n{INDENT*3}}}\n"
                    elif union_type == "DateTime":
                        class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.String)\n{INDENT*3}{{\n{INDENT*4}return new {union_class_name}(System.DateTime.Parse(element.GetString()));\n{INDENT*3}}}\n"
                    elif union_type == "DateTimeOffset":
                        class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.String)\n{INDENT*3}{{\n{INDENT*4}return new {union_class_name}(System.DateTimeOffset.Parse(element.GetString()));\n{INDENT*3}}}\n"
                else:
                    if union_type.startswith("global::"):
                        type_kind = self.generated_types[union_type] if union_type in self.generated_types else "class"
                        if type_kind == "class":
                            class_definition_read += f"{INDENT*3}if ({union_type}.IsJsonMatch(element))\n{INDENT*3}{{\n{INDENT*4}return new {union_class_name}({union_type}.FromData(element, System.Net.Mime.MediaTypeNames.Application.Json));\n{INDENT*3}}}\n"
                        elif type_kind == "enum":
                            class_definition_read += f"{INDENT*3}if (element.ValueKind == JsonValueKind.String && Enum.TryParse<{union_type}>(element.GetString(), true, out _ ))\n{INDENT*3}{{\n{INDENT*4}return new {union_class_name}(Enum.Parse<{union_type}>(element.GetString()));\n{INDENT*3}}}\n"
                class_definition_write += f"{INDENT*3}{'else ' if i>0 else ''}if (value.{union_type_name} != null)\n{INDENT*3}{{\n{INDENT*4}System.Text.Json.JsonSerializer.Serialize(writer, value.{union_type_name}, options);\n{INDENT*3}}}\n"
                gij = self.get_is_json_match_clause_type("element", class_name, union_type)
                if gij:
                    list_is_json_match.append(gij)

        class_definition = \
            f"/// <summary>\n/// {class_name}. Type union resolver. \n/// </summary>\n" + \
            f"public partial class {class_name}\n{{\n" + \
            f"{INDENT}/// <summary>\n{INDENT}/// Union class for {field_name}\n{INDENT}/// </summary>\n"
        if self.system_xml_annotation:
            class_definition += \
                f"{INDENT}[XmlRoot(\"{union_class_name}\")]\n"
        if self.msgpack_annotation:
            class_definition += \
                f"{INDENT}[MessagePackObject]\n"
        if self.system_text_json_annotation:
            class_definition += \
                f"{INDENT}[System.Text.Json.Serialization.JsonConverter(typeof({union_class_name}))]\n"
        if self.protobuf_net_annotation:
            class_definition += f"{INDENT}[ProtoContract]\n"
        class_definition += \
            f"{INDENT}public sealed class {union_class_name}"
        if self.system_text_json_annotation:
            class_definition += f": System.Text.Json.Serialization.JsonConverter<global::{namespace}.{class_name}.{union_class_name}>"
        class_definition += f"\n{INDENT}{{\n" + \
            f"{INDENT*2}/// <summary>\n{INDENT*2}/// Default constructor\n{INDENT*2}/// </summary>\n" + \
            f"{INDENT*2}public {union_class_name}() {{ }}\n"
        class_definition += class_definition_ctors
        if self.avro_annotation:
            class_definition += \
                f"{INDENT*2}/// <summary>\n{INDENT*2}/// Constructor for Avro decoder\n{INDENT*2}/// </summary>\n" + \
                f"{INDENT*2}internal static {union_class_name} FromObject(object obj)\n{INDENT*2}{{\n"
            if class_definition_genericrecordctor:
                class_definition += \
                    f"{INDENT*3}if (obj is global::Avro.Generic.GenericRecord)\n{INDENT*3}{{\n" + \
                    f"{INDENT*4}return new {union_class_name}((global::Avro.Generic.GenericRecord)obj);\n" + \
                    f"{INDENT*3}}}\n"
            class_definition += \
                f"{INDENT*3}var self = new {union_class_name}();\n" + \
                class_definition_objctr + \
                f"{INDENT*3}throw new NotSupportedException(\"No record type matched the type\");\n" + \
                f"{INDENT*2}}}\n"
            if class_definition_genericrecordctor:
                class_definition += f"\n{INDENT*2}/// <summary>\n{INDENT*2}/// Constructor from Avro GenericRecord\n{INDENT*2}/// </summary>\n" + \
                    f"{INDENT*2}public {union_class_name}(global::Avro.Generic.GenericRecord obj)\n{INDENT*2}{{\n" + \
                    class_definition_genericrecordctor + \
                    f"{INDENT*3}throw new NotSupportedException(\"No record type matched the type\");\n" + \
                    f"{INDENT*2}}}\n"
        class_definition += \
            class_definition_decls + \
            f"\n{INDENT*2}/// <summary>\n{INDENT*2}/// Yields the current value of the union\n{INDENT*2}/// </summary>\n" + \
            f"\n{INDENT*2}public Object ToObject()\n{INDENT*2}{{\n" + \
            class_definition_toobject+ \
            f"{INDENT*3}throw new NotSupportedException(\"No record type is set in the union\");\n" + \
            f"{INDENT*2}}}\n"
        if self.system_text_json_annotation:
            class_definition += \
                f"\n{INDENT*2}/// <summary>\n{INDENT*2}/// Reads the JSON representation of the object.\n{INDENT*2}/// </summary>\n" + \
                f"{INDENT*2}public override {union_class_name}? Read(ref Utf8JsonReader reader, System.Type typeToConvert, JsonSerializerOptions options)\n{INDENT*2}{{\n{INDENT*3}var element = JsonElement.ParseValue(ref reader);\n" + \
                class_definition_read + \
                f"{INDENT*3}throw new NotSupportedException(\"No record type matched the JSON data\");\n{INDENT*2}}}\n" + \
                f"\n{INDENT*2}/// <summary>\n{INDENT*2}/// Writes the JSON representation of the object.\n{INDENT*2}/// </summary>\n" + \
                f"{INDENT*2}public override void Write(Utf8JsonWriter writer, {union_class_name} value, JsonSerializerOptions options)\n{INDENT*2}{{\n" + \
                class_definition_write + \
                f"{INDENT*3}else\n{INDENT*3}{{\n{INDENT*4}throw new NotSupportedException(\"No record type is set in the union\");\n{INDENT*3}}}\n{INDENT*2}}}\n" + \
                f"\n{INDENT*2}/// <summary>\n{INDENT*2}/// Checks if the JSON element matches the schema\n{INDENT*2}/// </summary>\n" + \
                f"{INDENT*2}public static bool IsJsonMatch(System.Text.Json.JsonElement element)\n{INDENT*2}{{" + \
                f"\n{INDENT*3}return "+f"\n{INDENT*3} || ".join(list_is_json_match)+f";\n{INDENT*2}}}\n"
        
        # Generate Equals and GetHashCode for the union class
        # Build list of property names for comparison
        union_property_names = []
        for union_type in union_types:
            if union_type.startswith("Dictionary<"):
                match = re.findall(r"Dictionary<(.+)\s*,\s*(.+)>", union_type)
                union_property_names.append("Map" + pascal(match[0][1].rsplit('.', 1)[-1]))
            elif union_type.startswith("List<"):
                match = re.findall(r"List<(.+)>", union_type)
                union_property_names.append("Array" + pascal(match[0].rsplit('.', 1)[-1]))
            elif union_type == "byte[]":
                union_property_names.append("bytes")
            else:
                prop_name = union_type.rsplit('.', 1)[-1]
                if self.is_csharp_reserved_word(prop_name):
                    prop_name = f"@{prop_name}"
                union_property_names.append(prop_name)
        
        equals_conditions = " && ".join([f"Equals({name}, other.{name})" for name in union_property_names])
        
        # HashCode.Combine only accepts up to 8 arguments, so we need to chain calls for larger unions
        def generate_hashcode_expression(props: list) -> str:
            if len(props) <= 8:
                return f"HashCode.Combine({', '.join(props)})"
            else:
                # Chain HashCode.Combine calls: Combine(Combine(first 7...), next 7...)
                first_batch = props[:7]
                remaining = props[7:]
                inner = generate_hashcode_expression(remaining)
                return f"HashCode.Combine({', '.join(first_batch)}, {inner})"
        
        hashcode_expression = generate_hashcode_expression(union_property_names)
        
        class_definition += \
            f"\n{INDENT*2}/// <summary>\n{INDENT*2}/// Determines whether the specified object is equal to the current object.\n{INDENT*2}/// </summary>\n" + \
            f"{INDENT*2}public override bool Equals(object? obj)\n{INDENT*2}{{\n" + \
            f"{INDENT*3}if (obj is not {union_class_name} other) return false;\n" + \
            f"{INDENT*3}return {equals_conditions};\n" + \
            f"{INDENT*2}}}\n" + \
            f"\n{INDENT*2}/// <summary>\n{INDENT*2}/// Serves as the default hash function.\n{INDENT*2}/// </summary>\n" + \
            f"{INDENT*2}public override int GetHashCode()\n{INDENT*2}{{\n" + \
            f"{INDENT*3}return {hashcode_expression};\n" + \
            f"{INDENT*2}}}\n"
        
        class_definition += f"{INDENT}}}\n}}"

        if write_file:
            self.write_to_file(namespace, class_name +"."+union_class_name, class_definition)

        self.generated_types[ref] = "union" # it doesn't matter if the names clash, we just need to know whether it's a union
        return ref

    def find_type(self, kind: str, avro_schema: JsonNode, type_name: str, type_namespace: str, parent_namespace = '') -> JsonNode:
        """ recursively find the type (kind 'record' or 'enum') in the schema """
        if isinstance(avro_schema, list):
            for s in avro_schema:
                found = self.find_type(kind, s, type_name, type_namespace, parent_namespace)
                if found:
                    return found
        elif isinstance(avro_schema, dict):
            if avro_schema.get('type') == kind and avro_schema.get('name') == type_name and avro_schema.get('namespace', parent_namespace) == type_namespace:
                return avro_schema
            parent_namespace = avro_schema.get('namespace', parent_namespace)
            if 'fields' in avro_schema and isinstance(avro_schema['fields'], list):
                for field in avro_schema['fields']:
                    if isinstance(field, dict) and 'type' in field:
                        # Recursively search within field types (including union arrays)
                        found = self.find_type(kind, field['type'], type_name, type_namespace, parent_namespace)
                        if found:
                            return found
        return None

    def is_enum_type(self, avro_type: Union[str, Dict, List]) -> bool:
        """ Checks if a type is an enum (including nullable enums) """
        if isinstance(avro_type, str):
            if avro_type in ('null', 'boolean', 'int', 'long', 'float', 'double', 'bytes', 'string'):
                return False
            schema = self.schema_doc
            name = avro_type.split('.')[-1]
            namespace = ".".join(avro_type.split('.')[:-1])
            return self.find_type('enum', schema, name, namespace) is not None
        elif isinstance(avro_type, list):
            # Check for nullable enum: ["null", <enum-type>] or [<enum-type>, "null"]
            non_null_types = [t for t in avro_type if t != 'null']
            if len(non_null_types) == 1:
                return self.is_enum_type(non_null_types[0])
            return False
        elif isinstance(avro_type, dict):
            return avro_type.get('type') == 'enum'
        return False

    def generate_property(self, field_index: int, field: Dict, class_name: str, parent_namespace: str) -> str:
        """ Generates a property """
        is_enum_type = self.is_enum_type(field['type'])
        field_type = self.convert_avro_type_to_csharp(
            class_name, field['name'], field['type'], parent_namespace)
        field_default = field.get('const', field.get('default', None))
        annotation_name = field_name = field['name']
        if self.is_csharp_reserved_word(field_name):
            field_name = f"@{field_name}"
        if self.pascal_properties:
            field_name = pascal(field_name)
        if field_name == class_name:
            field_name += "_"
        prop = ''
        prop += f"{INDENT}/// <summary>\n{INDENT}/// { field.get('doc', field_name) }\n{INDENT}/// </summary>\n"
        
        if self.protobuf_net_annotation:
            prop += f"{INDENT}[ProtoMember({field_index}, Name=\"{annotation_name}\")]\n"
        
        # Add XML serialization attribute if enabled
        if self.system_xml_annotation:
            xmlkind = field.get('xmlkind', 'element')
            if xmlkind == 'element':
                prop += f"{INDENT}[XmlElement(\"{annotation_name}\")]\n"
            elif xmlkind == 'attribute':
                prop += f"{INDENT}[XmlAttribute(\"{annotation_name}\")]\n"
        
        # Add MessagePack serialization attribute if enabled
        if self.msgpack_annotation:
            prop += f"{INDENT}[Key({field_index})]\n"

        if self.system_text_json_annotation:
            prop += f"{INDENT}[System.Text.Json.Serialization.JsonPropertyName(\"{annotation_name}\")]\n"
            if is_enum_type:
                prop += f"{INDENT}[System.Text.Json.Serialization.JsonConverter(typeof(JsonStringEnumConverter))]\n"
            if field_type.endswith("Union") and not field_type.startswith("global::"):
                prop += f"{INDENT}[System.Text.Json.Serialization.JsonConverter(typeof({field_type}))]\n"
        if self.newtonsoft_json_annotation:
            prop += f"{INDENT}[Newtonsoft.Json.JsonProperty(\"{annotation_name}\")]\n"
            if is_enum_type:
                prop += f"{INDENT}[Newtonsoft.Json.JsonConverter(typeof(Newtonsoft.Json.Converters.StringEnumConverter))]\n"
        
        # Determine initialization value
        initialization = ""
        if field_default is not None:
            # Has explicit default value
            if is_enum_type:
                # For enum types, use qualified enum value (e.g., Type.Circle)
                # Get the base enum type name (strip nullable ? suffix if present)
                enum_type = field_type.rstrip('?')
                initialization = f" = {enum_type}.{field_default};"
            elif isinstance(field_default, str):
                initialization = f" = \"{field_default}\";"
            else:
                initialization = f" = {field_default};"
        elif field_type == "string":
            # Non-nullable string without default should be initialized to empty string
            initialization = " = string.Empty;"
        elif not field_type.endswith("?") and field_type.startswith("List<"):
            # Non-nullable List should be initialized to empty list
            initialization = " = new();"
        elif not field_type.endswith("?") and field_type.startswith("Dictionary<"):
            # Non-nullable Dictionary should be initialized to empty dictionary
            initialization = " = new();"
        elif not field_type.endswith("?") and field_type.startswith("global::") and not self.is_csharp_primitive_type(field_type):
            # Non-nullable custom reference types should be initialized with new instance
            initialization = " = new();"
        
        prop += f"{INDENT}public {field_type} {field_name} {{ get; set; }}{initialization}"
        return prop

    def write_to_file(self, namespace: str, name: str, definition: str):
        """ Writes the class or enum to a file """
        directory_path = os.path.join(
            self.output_dir, os.path.join('src', namespace.replace('.', os.sep)))
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)
        file_path = os.path.join(directory_path, f"{name}.cs")

        with open(file_path, 'w', encoding='utf-8') as file:
            # Common using statements (add more as needed)
            file_content = "using System;\nusing System.Collections.Generic;\n"
            file_content += "using System.Linq;\n"
            if self.protobuf_net_annotation:
                file_content += "using ProtoBuf;\n"
            if self.system_text_json_annotation:
                file_content += "using System.Text.Json;\n"
                file_content += "using System.Text.Json.Serialization;\n"
            if self.newtonsoft_json_annotation:
                file_content += "using Newtonsoft.Json;\n"
            if self.system_xml_annotation:  # Add XML serialization using directive
                file_content += "using System.Xml.Serialization;\n"
            if self.msgpack_annotation:  # Add MessagePack serialization using directive
                file_content += "using MessagePack;\n"

            if namespace:
                # Namespace declaration with correct indentation for the definition
                file_content += f"\nnamespace {namespace}\n{{\n"
                indented_definition = '\n'.join(
                    [f"{INDENT}{line}" for line in definition.split('\n')])
                file_content += f"{indented_definition}\n}}"
            else:
                file_content += definition
            file.write(file_content)

    def generate_tests(self, output_dir: str) -> None:
        """ Generates unit tests for all the generated C# classes and enums """
        test_directory_path = os.path.join(output_dir, "test")
        if not os.path.exists(test_directory_path):
            os.makedirs(test_directory_path, exist_ok=True)

        for class_name, type_kind in self.generated_types.items():
            if type_kind in ["class", "enum"]:
                self.generate_test_class(class_name, type_kind, test_directory_path)

    def generate_test_class(self, class_name: str, type_kind: str, test_directory_path: str) -> None:
        """ Generates a unit test class for a given C# class or enum """
        avro_schema:Dict[str,JsonNode] = cast(Dict[str,JsonNode], self.generated_avro_types.get(class_name, {}))
        if class_name.startswith("global::"):
            class_name = class_name[8:]
        test_class_name = f"{class_name.split('.')[-1]}Tests"
        namespace = ".".join(class_name.split('.')[:-1])
        class_base_name = class_name.split('.')[-1]

        if type_kind == "class":
            fields = self.get_class_test_fields(avro_schema, class_base_name)
            test_class_definition = process_template(
                "avrotocsharp/class_test.cs.jinja",
                namespace=namespace,
                test_class_name=test_class_name,
                class_base_name=class_base_name,
                fields=fields,
                avro_annotation=self.avro_annotation,
                system_xml_annotation=self.system_xml_annotation,
                system_text_json_annotation=self.system_text_json_annotation,
                newtonsoft_json_annotation=self.newtonsoft_json_annotation,
                protobuf_net_annotation=self.protobuf_net_annotation,
                msgpack_annotation=self.msgpack_annotation
            )
        elif type_kind == "enum":
            test_class_definition = process_template(
                "avrotocsharp/enum_test.cs.jinja",
                namespace=namespace,
                test_class_name=test_class_name,
                enum_base_name=class_base_name,
                symbols=avro_schema.get('symbols', []),
                avro_annotation=self.avro_annotation,
                system_xml_annotation=self.system_xml_annotation,
                system_text_json_annotation=self.system_text_json_annotation,
                newtonsoft_json_annotation=self.newtonsoft_json_annotation,
                protobuf_net_annotation=self.protobuf_net_annotation
            )

        test_file_path = os.path.join(test_directory_path, f"{test_class_name}.cs")
        with open(test_file_path, 'w', encoding='utf-8') as test_file:
            test_file.write(test_class_definition)

    def get_class_test_fields(self, avro_schema: Dict[str,JsonNode], class_name: str) -> List[Any]:
        """ Retrieves fields for a given class name """

        class Field:
            def __init__(self, fn: str, ft:str, tv:Any, ct: bool, pm: bool, ie: bool, iu: bool):
                self.field_name = fn
                self.field_type = ft
                self.test_value = tv
                self.is_const = ct
                self.is_primitive = pm
                self.is_enum = ie
                self.is_union = iu

        fields: List[Field] = []
        if avro_schema and 'fields' in avro_schema:
            for field in cast(List[Dict[str,JsonNode]],avro_schema['fields']):
                field_name = str(field['name'])
                if self.pascal_properties:
                    field_name = pascal(field_name)
                if field_name == class_name:
                    field_name += "_"
                if self.is_csharp_reserved_word(field_name):
                    field_name = f"@{field_name}"
                field_type = self.convert_avro_type_to_csharp(class_name, field_name, field['type'], str(avro_schema.get('namespace', '')))
                is_class = field_type in self.generated_types and self.generated_types[field_type] == "class"
                is_enum = self.is_enum_type(field['type'])
                is_union = field_type in self.generated_types and self.generated_types[field_type] == "union"
                test_value = self.get_test_value(field_type, field['type'], str(avro_schema.get('namespace', ''))) if not "const" in field else '\"'+str(field["const"])+'\"'
                f = Field(field_name,
                          field_type,
                          test_value,
                          "const" in field and field["const"] is not None,
                          not is_class,
                          is_enum,
                          is_union)
                fields.append(f)
        return cast(List[Any], fields)

    def get_test_value(self, csharp_type: str, avro_type: JsonNode = None, parent_namespace: str = '') -> str:
        """Returns a default test value based on the Avro type"""
        # For nullable object types, return typed null to avoid var issues
        if csharp_type == "object?":
            return "(object?)null"
        
        test_values = {
            'string': '"test_string"',
            'bool': 'true',
            'int': '42',
            'long': '42L',
            'float': '3.14f',
            'double': '3.14',
            'decimal': '3.14d',
            'byte[]': 'new byte[] { 0x01, 0x02, 0x03 }',
            'null': 'null',
            'Date': 'new Date()',
            'DateTime': 'DateTime.UtcNow()',
            'Guid': 'Guid.NewGuid()'
        }
        if csharp_type.endswith('?'):
            csharp_type = csharp_type[:-1]
        
        # Check if this is a union type (either by generated_types lookup or by type structure)
        if csharp_type in self.generated_types and self.generated_types[csharp_type] == "union":
            # For union types, we need to initialize with one of the valid options
            # Find the first non-null type in the union and create a value for it
            if isinstance(avro_type, list):
                non_null_types = [t for t in avro_type if t != 'null']
                if non_null_types:
                    first_option = non_null_types[0]
                    first_option_csharp = self.convert_avro_type_to_csharp('', 'Option0', first_option, parent_namespace)
                    first_option_value = self.get_test_value(first_option_csharp, first_option, parent_namespace)
                    return f'new {csharp_type}({first_option_value})'
        
        return test_values.get(csharp_type, f'new {csharp_type}()')

    def convert_schema(self, schema: JsonNode, output_dir: str):
        """ Converts Avro schema to C# """
        if not isinstance(schema, list):
            schema = [schema]

        # Determine project name: use explicit project_name if set, otherwise derive from base_namespace
        if self.project_name and self.project_name.strip():
            # Use explicitly set project name
            project_name = self.project_name
        else:
            # Fall back to using base_namespace as project name
            project_name = self.base_namespace
            if not project_name or project_name.strip() == '':
                # Derive from output directory name as fallback
                project_name = os.path.basename(os.path.abspath(output_dir))
                if not project_name or project_name.strip() == '':
                    project_name = 'Generated'
                # Clean up the project name
                project_name = project_name.replace('-', '_').replace(' ', '_')
                # Update base_namespace to match (only if it was empty)
                self.base_namespace = project_name
                import warnings
                warnings.warn(f"No namespace provided, using '{project_name}' derived from output directory", UserWarning)
        
        self.schema_doc = schema
        self.type_dict = build_flat_type_dict(self.schema_doc)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        if not glob.glob(os.path.join(output_dir, "src", "*.sln")):
            sln_file = os.path.join(
                output_dir, f"{project_name}.sln")
            if not os.path.exists(sln_file):
                if not os.path.exists(os.path.dirname(sln_file)):
                    os.makedirs(os.path.dirname(sln_file))
                with open(sln_file, 'w', encoding='utf-8') as file:
                    file.write(process_template(
                        "avrotocsharp/project.sln.jinja", 
                        project_name=project_name, 
                        uuid=lambda:str(uuid.uuid4()),
                        avro_annotation=self.avro_annotation,
                        system_xml_annotation=self.system_xml_annotation,
                        system_text_json_annotation=self.system_text_json_annotation,
                        newtonsoft_json_annotation=self.newtonsoft_json_annotation))
        if not glob.glob(os.path.join(output_dir, "src", "*.csproj")):
            csproj_file = os.path.join(
                output_dir, "src", f"{pascal(project_name)}.csproj")
            if not os.path.exists(csproj_file):
                if not os.path.exists(os.path.dirname(csproj_file)):
                    os.makedirs(os.path.dirname(csproj_file))
                with open(csproj_file, 'w', encoding='utf-8') as file:
                    file.write(process_template(
                        "avrotocsharp/project.csproj.jinja",
                        project_name=project_name, 
                        avro_annotation=self.avro_annotation,
                        system_xml_annotation=self.system_xml_annotation,
                        system_text_json_annotation=self.system_text_json_annotation,
                        newtonsoft_json_annotation=self.newtonsoft_json_annotation,
                        protobuf_net_annotation=self.protobuf_net_annotation,
                        msgpack_annotation=self.msgpack_annotation,
                        CSHARP_AVRO_VERSION=CSHARP_AVRO_VERSION,
                        NEWTONSOFT_JSON_VERSION=NEWTONSOFT_JSON_VERSION,
                        SYSTEM_TEXT_JSON_VERSION=SYSTEM_TEXT_JSON_VERSION,
                        SYSTEM_MEMORY_DATA_VERSION=SYSTEM_MEMORY_DATA_VERSION,
                        PROTOBUF_NET_VERSION=PROTOBUF_NET_VERSION,
                        MSGPACK_VERSION=MSGPACK_VERSION,
                        NUNIT_VERSION=NUNIT_VERSION,
                        NUNIT_ADAPTER_VERSION=NUNIT_ADAPTER_VERSION,
                        MSTEST_SDK_VERSION=MSTEST_SDK_VERSION))
        if not glob.glob(os.path.join(output_dir, "test", "*.csproj")):
            csproj_test_file = os.path.join(
                output_dir, "test", f"{pascal(project_name)}.Test.csproj")
            if not os.path.exists(csproj_test_file):
                if not os.path.exists(os.path.dirname(csproj_test_file)):
                    os.makedirs(os.path.dirname(csproj_test_file))
                with open(csproj_test_file, 'w', encoding='utf-8') as file:
                    file.write(process_template(
                        "avrotocsharp/testproject.csproj.jinja", 
                        project_name=project_name,
                        avro_annotation=self.avro_annotation,
                        system_xml_annotation=self.system_xml_annotation,
                        system_text_json_annotation=self.system_text_json_annotation,
                        newtonsoft_json_annotation=self.newtonsoft_json_annotation,
                        protobuf_net_annotation=self.protobuf_net_annotation,
                        NUNIT_VERSION=NUNIT_VERSION,
                        NUNIT_ADAPTER_VERSION=NUNIT_ADAPTER_VERSION,
                        MSTEST_SDK_VERSION=MSTEST_SDK_VERSION,
                        COVERLET_VERSION=COVERLET_VERSION))

        # Generate coverage scripts
        if not os.path.exists(os.path.join(output_dir, "run_coverage.ps1")):
            coverage_ps1_file = os.path.join(output_dir, "run_coverage.ps1")
            with open(coverage_ps1_file, 'w', encoding='utf-8') as file:
                file.write(process_template(
                    "avrotocsharp/run_coverage.ps1.jinja", 
                    project_name=project_name))
        
        if not os.path.exists(os.path.join(output_dir, "run_coverage.sh")):
            coverage_sh_file = os.path.join(output_dir, "run_coverage.sh")
            with open(coverage_sh_file, 'w', encoding='utf-8') as file:
                file.write(process_template(
                    "avrotocsharp/run_coverage.sh.jinja", 
                    project_name=project_name))
            # Make the shell script executable on Unix-like systems
            try:
                os.chmod(coverage_sh_file, 0o755)
            except:
                pass  # Ignore on Windows
        
        # Generate README with coverage documentation
        if not os.path.exists(os.path.join(output_dir, "README.md")):
            readme_file = os.path.join(output_dir, "README.md")
            with open(readme_file, 'w', encoding='utf-8') as file:
                file.write(process_template(
                    "avrotocsharp/README.md.jinja", 
                    project_name=project_name))

        self.output_dir = output_dir
        for avro_schema in (avs for avs in schema if isinstance(avs, dict)):
            self.generate_class_or_enum(avro_schema, '')
        self.generate_tests(output_dir)

    def convert(self, avro_schema_path: str, output_dir: str):
        """ Converts Avro schema to C# """
        with open(avro_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        self.convert_schema(schema, output_dir)


def convert_avro_to_csharp(
    avro_schema_path, 
    cs_file_path, 
    base_namespace='',
    project_name='',
    pascal_properties=False, 
    system_text_json_annotation=False, 
    newtonsoft_json_annotation=False, 
    system_xml_annotation=False,
    msgpack_annotation=False,
    avro_annotation=False,
    protobuf_net_annotation=False
):
    """Converts Avro schema to C# classes

    Args:
        avro_schema_path (str): Avro input schema path
        cs_file_path (str): Output C# file path
        base_namespace (str, optional): Base namespace. Defaults to ''.
        project_name (str, optional): Explicit project name for .csproj files (separate from namespace). Defaults to ''.
        pascal_properties (bool, optional): Pascal case properties. Defaults to False.
        system_text_json_annotation (bool, optional): Use System.Text.Json annotations. Defaults to False.
        newtonsoft_json_annotation (bool, optional): Use Newtonsoft.Json annotations. Defaults to False.
        system_xml_annotation (bool, optional): Use System.Xml.Serialization annotations. Defaults to False.
        msgpack_annotation (bool, optional): Use MessagePack annotations. Defaults to False.
        avro_annotation (bool, optional): Use Avro annotations. Defaults to False.
        protobuf_net_annotation (bool, optional): Use protobuf-net annotations. Defaults to False.
    """

    if not base_namespace:
        base_namespace = os.path.splitext(os.path.basename(cs_file_path))[0].replace('-', '_')
    avrotocs = AvroToCSharp(base_namespace)
    avrotocs.project_name = project_name
    avrotocs.pascal_properties = pascal_properties
    avrotocs.system_text_json_annotation = system_text_json_annotation
    avrotocs.newtonsoft_json_annotation = newtonsoft_json_annotation
    avrotocs.system_xml_annotation = system_xml_annotation
    avrotocs.msgpack_annotation = msgpack_annotation
    avrotocs.avro_annotation = avro_annotation
    avrotocs.protobuf_net_annotation = protobuf_net_annotation
    avrotocs.convert(avro_schema_path, cs_file_path)


def convert_avro_schema_to_csharp(
    avro_schema: JsonNode, 
    output_dir: str, 
    base_namespace: str = '',
    project_name: str = '',
    pascal_properties: bool = False, 
    system_text_json_annotation: bool = False, 
    newtonsoft_json_annotation: bool = False, 
    system_xml_annotation: bool = False,
    msgpack_annotation: bool = False,
    avro_annotation: bool = False,
    protobuf_net_annotation: bool = False
):
    """Converts Avro schema to C# classes

    Args:
        avro_schema (JsonNode): Avro schema to convert
        output_dir (str): Output directory
        base_namespace (str, optional): Base namespace for the generated classes. Defaults to ''.
        project_name (str, optional): Explicit project name for .csproj files (separate from namespace). Defaults to ''.
        pascal_properties (bool, optional): Pascal case properties. Defaults to False.
        system_text_json_annotation (bool, optional): Use System.Text.Json annotations. Defaults to False.
        newtonsoft_json_annotation (bool, optional): Use Newtonsoft.Json annotations. Defaults to False.
        system_xml_annotation (bool, optional): Use System.Xml.Serialization annotations. Defaults to False.
        msgpack_annotation (bool, optional): Use MessagePack annotations. Defaults to False.
        avro_annotation (bool, optional): Use Avro annotations. Defaults to False.
        protobuf_net_annotation (bool, optional): Use protobuf-net annotations. Defaults to False.
    """
    avrotocs = AvroToCSharp(base_namespace)
    avrotocs.project_name = project_name
    avrotocs.pascal_properties = pascal_properties
    avrotocs.system_text_json_annotation = system_text_json_annotation
    avrotocs.newtonsoft_json_annotation = newtonsoft_json_annotation
    avrotocs.system_xml_annotation = system_xml_annotation
    avrotocs.msgpack_annotation = msgpack_annotation
    avrotocs.avro_annotation = avro_annotation
    avrotocs.protobuf_net_annotation = protobuf_net_annotation
    avrotocs.convert_schema(avro_schema, output_dir)



"""
Central dependency version resolver for Avrotize code generation templates.

This module provides functions to read dependency versions from central project files
located in avrotize/dependencies/{language}/{runtime_version}/. These files are
monitored by Dependabot, which proposes updates when new versions are available.

Supported languages:
- cs: C# (.NET) - reads from dependencies.csproj
- java: Java - reads from pom.xml
- py: Python - reads from requirements.txt
- ts: TypeScript - reads from package.json
- go: Go - reads from go.mod
- rust: Rust - reads from Cargo.toml
- cpp: C++ - reads from vcpkg.json
"""

import json
import os
import re
import xml.etree.ElementTree as ET
from typing import Optional


def get_dependency(language: str, runtime_version: str, dependency_name: str) -> str:
    """
    Get the dependency declaration for a specific language and dependency.
    
    Args:
        language: Language identifier ('cs', 'java', 'py/python', 'ts/typescript', 'go', 'rust', 'cpp')
        runtime_version: Runtime version identifier (e.g., 'net90', 'jdk21', 'py312')
        dependency_name: The dependency identifier. For Java, can use 'groupId:artifactId' 
                        syntax to disambiguate (e.g., 'org.junit.jupiter:junit-jupiter')
    
    Returns:
        The dependency declaration string in the appropriate format for the language.
        
    Raises:
        ValueError: If the language is unsupported or dependency is not found.
    """
    # Determine the path to the central dependency file
    deps_dir = os.path.join(os.path.dirname(__file__), 'dependencies')
    
    # Normalize language aliases
    lang = language.lower()
    if lang in ('python', 'py'):
        lang = 'py'
    elif lang in ('typescript', 'ts'):
        lang = 'ts'
    elif lang in ('csharp', 'c#'):
        lang = 'cs'
    
    if lang == 'cs':
        master_project_path = os.path.join(deps_dir, f'cs/{runtime_version}/dependencies.csproj')
    elif lang == 'java':
        master_project_path = os.path.join(deps_dir, f'java/{runtime_version}/pom.xml')
    elif lang == 'py':
        master_project_path = os.path.join(deps_dir, f'python/{runtime_version}/requirements.txt')
    elif lang == 'ts':
        master_project_path = os.path.join(deps_dir, f'typescript/{runtime_version}/package.json')
    elif lang == 'go':
        master_project_path = os.path.join(deps_dir, f'go/{runtime_version}/go.mod')
    elif lang == 'rust':
        master_project_path = os.path.join(deps_dir, f'rust/{runtime_version}/Cargo.toml')
    elif lang == 'cpp':
        master_project_path = os.path.join(deps_dir, f'cpp/{runtime_version}/vcpkg.json')
    else:
        raise ValueError(f"Unsupported language: {language}")
    
    if not os.path.exists(master_project_path):
        raise ValueError(f"Dependency file not found: {master_project_path}")
    
    if lang == 'cs':
        return _get_nuget_dependency(master_project_path, dependency_name)
    elif lang == 'java':
        return _get_maven_dependency(master_project_path, dependency_name)
    elif lang == 'py':
        return _get_python_dependency(master_project_path, dependency_name)
    elif lang == 'ts':
        return _get_npm_dependency(master_project_path, dependency_name)
    elif lang == 'go':
        return _get_go_dependency(master_project_path, dependency_name)
    elif lang == 'rust':
        return _get_cargo_dependency(master_project_path, dependency_name)
    elif lang == 'cpp':
        return _get_vcpkg_dependency(master_project_path, dependency_name)
    
    raise ValueError(f"Unsupported language: {language}")


def get_dependency_version(language: str, runtime_version: str, dependency_name: str) -> str:
    """
    Get just the version string for a dependency.
    
    Args:
        language: Language identifier (cs, java, py/python, ts/typescript, go, rust, cpp)
        runtime_version: Runtime version identifier
        dependency_name: The dependency identifier
        
    Returns:
        The version string only (e.g., '1.12.0', '2.18.2')
    """
    deps_dir = os.path.join(os.path.dirname(__file__), 'dependencies')
    
    # Normalize language aliases
    lang = language.lower()
    if lang in ('python', 'py'):
        lang = 'py'
    elif lang in ('typescript', 'ts'):
        lang = 'ts'
    elif lang in ('csharp', 'c#'):
        lang = 'cs'
    
    if lang == 'cs':
        master_project_path = os.path.join(deps_dir, f'cs/{runtime_version}/dependencies.csproj')
        return _get_nuget_version(master_project_path, dependency_name)
    elif lang == 'java':
        master_project_path = os.path.join(deps_dir, f'java/{runtime_version}/pom.xml')
        return _get_maven_version(master_project_path, dependency_name)
    elif lang == 'py':
        master_project_path = os.path.join(deps_dir, f'python/{runtime_version}/requirements.txt')
        return _get_python_version(master_project_path, dependency_name)
    elif lang == 'ts':
        master_project_path = os.path.join(deps_dir, f'typescript/{runtime_version}/package.json')
        return _get_npm_version(master_project_path, dependency_name)
    elif lang == 'go':
        master_project_path = os.path.join(deps_dir, f'go/{runtime_version}/go.mod')
        return _get_go_version(master_project_path, dependency_name)
    elif lang == 'rust':
        master_project_path = os.path.join(deps_dir, f'rust/{runtime_version}/Cargo.toml')
        return _get_cargo_version(master_project_path, dependency_name)
    elif lang == 'cpp':
        master_project_path = os.path.join(deps_dir, f'cpp/{runtime_version}/vcpkg.json')
        return _get_vcpkg_version(master_project_path, dependency_name)
    
    raise ValueError(f"Unsupported language: {language}")


def _get_nuget_dependency(project_path: str, package_name: str) -> str:
    """Get NuGet PackageReference XML element."""
    tree = ET.parse(project_path)
    root = tree.getroot()
    
    for package in root.findall(".//PackageReference"):
        if package.get('Include') == package_name:
            return ET.tostring(package, encoding='unicode').strip()
    
    raise ValueError(f"Dependency '{package_name}' not found in {project_path}")


def _get_nuget_version(project_path: str, package_name: str) -> str:
    """Get NuGet package version."""
    tree = ET.parse(project_path)
    root = tree.getroot()
    
    for package in root.findall(".//PackageReference"):
        if package.get('Include') == package_name:
            return package.get('Version', '')
    
    raise ValueError(f"Dependency '{package_name}' not found in {project_path}")


def _get_maven_dependency(project_path: str, dependency_name: str) -> str:
    """Get Maven dependency XML element."""
    tree = ET.parse(project_path)
    root = tree.getroot()
    
    # Support groupId:artifactId syntax for disambiguation
    if ':' in dependency_name:
        target_group_id, target_artifact_id = dependency_name.split(':', 1)
    else:
        target_group_id = None
        target_artifact_id = dependency_name
    
    for dependency in root.findall(".//dependency"):
        artifact_id = dependency.find('artifactId')
        group_id = dependency.find('groupId')
        
        if artifact_id is not None and artifact_id.text == target_artifact_id:
            # If groupId was specified, verify it matches
            if target_group_id is not None:
                if group_id is None or group_id.text != target_group_id:
                    continue
            return ET.tostring(dependency, encoding='unicode').strip()
    
    raise ValueError(f"Dependency '{dependency_name}' not found in {project_path}")


def _get_maven_version(project_path: str, dependency_name: str) -> str:
    """Get Maven dependency version."""
    tree = ET.parse(project_path)
    root = tree.getroot()
    
    if ':' in dependency_name:
        target_group_id, target_artifact_id = dependency_name.split(':', 1)
    else:
        target_group_id = None
        target_artifact_id = dependency_name
    
    for dependency in root.findall(".//dependency"):
        artifact_id = dependency.find('artifactId')
        group_id = dependency.find('groupId')
        
        if artifact_id is not None and artifact_id.text == target_artifact_id:
            if target_group_id is not None:
                if group_id is None or group_id.text != target_group_id:
                    continue
            version_elem = dependency.find('version')
            if version_elem is not None and version_elem.text:
                return version_elem.text
    
    raise ValueError(f"Dependency '{dependency_name}' not found in {project_path}")


def _get_python_dependency(project_path: str, package_name: str) -> str:
    """Get Python dependency from requirements.txt."""
    with open(project_path, 'r', encoding='utf-8') as f:
        content = f.read()
    
    for line in content.split('\n'):
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        
        # Match package name with version specifier
        match = re.match(r'^([a-zA-Z0-9_-]+)\s*([>=<!=]+.*)$', line)
        if match:
            pkg_name = match.group(1)
            version_spec = match.group(2)
            if pkg_name == package_name:
                return f"{pkg_name}{version_spec}"
    
    raise ValueError(f"Dependency '{package_name}' not found in {project_path}")


def _get_python_version(project_path: str, package_name: str) -> str:
    """Get Python package version."""
    with open(project_path, 'r', encoding='utf-8') as f:
        content = f.read()
    
    for line in content.split('\n'):
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        
        match = re.match(r'^([a-zA-Z0-9_-]+)\s*[>=<!=]+\s*(.+)$', line)
        if match:
            pkg_name = match.group(1)
            version = match.group(2).strip()
            if pkg_name == package_name:
                return version
    
    raise ValueError(f"Dependency '{package_name}' not found in {project_path}")


def _get_npm_dependency(project_path: str, package_name: str) -> str:
    """Get npm dependency as JSON key-value pair."""
    with open(project_path, 'r', encoding='utf-8') as f:
        package_data = json.load(f)
    
    dependencies = package_data.get('dependencies', {})
    dev_dependencies = package_data.get('devDependencies', {})
    
    if package_name in dependencies:
        version = dependencies[package_name]
        return f'"{package_name}": "{version}"'
    
    if package_name in dev_dependencies:
        version = dev_dependencies[package_name]
        return f'"{package_name}": "{version}"'
    
    raise ValueError(f"Dependency '{package_name}' not found in {project_path}")


def _get_npm_version(project_path: str, package_name: str) -> str:
    """Get npm package version."""
    with open(project_path, 'r', encoding='utf-8') as f:
        package_data = json.load(f)
    
    dependencies = package_data.get('dependencies', {})
    dev_dependencies = package_data.get('devDependencies', {})
    
    if package_name in dependencies:
        return dependencies[package_name].lstrip('^~')
    
    if package_name in dev_dependencies:
        return dev_dependencies[package_name].lstrip('^~')
    
    raise ValueError(f"Dependency '{package_name}' not found in {project_path}")


def _get_go_dependency(project_path: str, module_name: str) -> str:
    """Get Go module dependency line."""
    with open(project_path, 'r', encoding='utf-8') as f:
        content = f.read()
    
    # Parse require block
    require_pattern = re.compile(r'require\s*\(\s*([\s\S]*?)\s*\)', re.MULTILINE)
    require_match = require_pattern.search(content)
    
    if require_match:
        require_block = require_match.group(1)
        for line in require_block.strip().split('\n'):
            line = line.strip()
            if not line or line.startswith('//'):
                continue
            parts = line.split()
            if len(parts) >= 2:
                mod_path = parts[0]
                version = parts[1]
                if mod_path == module_name:
                    return f"\t{mod_path} {version}"
    
    raise ValueError(f"Dependency '{module_name}' not found in {project_path}")


def _get_go_version(project_path: str, module_name: str) -> str:
    """Get Go module version."""
    with open(project_path, 'r', encoding='utf-8') as f:
        content = f.read()
    
    require_pattern = re.compile(r'require\s*\(\s*([\s\S]*?)\s*\)', re.MULTILINE)
    require_match = require_pattern.search(content)
    
    if require_match:
        require_block = require_match.group(1)
        for line in require_block.strip().split('\n'):
            line = line.strip()
            if not line or line.startswith('//'):
                continue
            parts = line.split()
            if len(parts) >= 2:
                mod_path = parts[0]
                version = parts[1]
                if mod_path == module_name:
                    return version.lstrip('v')
    
    raise ValueError(f"Dependency '{module_name}' not found in {project_path}")


def _get_cargo_dependency(project_path: str, crate_name: str) -> str:
    """Get Rust/Cargo dependency as TOML line."""
    with open(project_path, 'r', encoding='utf-8') as f:
        content = f.read()
    
    # Look for dependencies in [dependencies] section
    in_deps_section = False
    for line in content.split('\n'):
        line_stripped = line.strip()
        
        if line_stripped == '[dependencies]':
            in_deps_section = True
            continue
        elif line_stripped.startswith('[') and line_stripped.endswith(']'):
            in_deps_section = False
            continue
        
        if in_deps_section and line_stripped:
            # Match simple version: crate = "version"
            simple_match = re.match(rf'^{re.escape(crate_name)}\s*=\s*"([^"]+)"', line_stripped)
            if simple_match:
                return line_stripped
            
            # Match complex version: crate = { version = "...", ... }
            complex_match = re.match(rf'^{re.escape(crate_name)}\s*=\s*\{{.*\}}', line_stripped)
            if complex_match:
                return line_stripped
    
    raise ValueError(f"Dependency '{crate_name}' not found in {project_path}")


def _get_cargo_version(project_path: str, crate_name: str) -> str:
    """Get Rust/Cargo crate version."""
    with open(project_path, 'r', encoding='utf-8') as f:
        content = f.read()
    
    in_deps_section = False
    for line in content.split('\n'):
        line_stripped = line.strip()
        
        if line_stripped == '[dependencies]':
            in_deps_section = True
            continue
        elif line_stripped.startswith('[') and line_stripped.endswith(']'):
            in_deps_section = False
            continue
        
        if in_deps_section and line_stripped:
            # Match simple version
            simple_match = re.match(rf'^{re.escape(crate_name)}\s*=\s*"([^"]+)"', line_stripped)
            if simple_match:
                return simple_match.group(1)
            
            # Match complex version with version key
            version_match = re.search(rf'^{re.escape(crate_name)}\s*=\s*\{{.*version\s*=\s*"([^"]+)"', line_stripped)
            if version_match:
                return version_match.group(1)
    
    raise ValueError(f"Dependency '{crate_name}' not found in {project_path}")


def _get_vcpkg_dependency(project_path: str, package_name: str) -> str:
    """Get vcpkg dependency as JSON object."""
    with open(project_path, 'r', encoding='utf-8') as f:
        vcpkg_data = json.load(f)
    
    dependencies = vcpkg_data.get('dependencies', [])
    for dep in dependencies:
        if isinstance(dep, dict):
            if dep.get('name') == package_name:
                return json.dumps(dep, indent=2)
        elif isinstance(dep, str) and dep == package_name:
            return f'"{package_name}"'
    
    raise ValueError(f"Dependency '{package_name}' not found in {project_path}")


def _get_vcpkg_version(project_path: str, package_name: str) -> str:
    """Get vcpkg package version."""
    with open(project_path, 'r', encoding='utf-8') as f:
        vcpkg_data = json.load(f)
    
    dependencies = vcpkg_data.get('dependencies', [])
    for dep in dependencies:
        if isinstance(dep, dict):
            if dep.get('name') == package_name:
                # Try various version keys
                for version_key in ['version>=', 'version', 'version>']:
                    if version_key in dep:
                        return dep[version_key]
    
    raise ValueError(f"Dependency '{package_name}' not found in {project_path}")



"""Convert a JSON Structure schema to an Iceberg schema."""

import json
import sys
from typing import Dict, List, Any, Optional
import pyarrow as pa
from pyiceberg.schema import Schema, NestedField
from pyiceberg.types import (
    BooleanType,
    IntegerType,
    LongType,
    FloatType,
    DoubleType,
    StringType,
    BinaryType,
    DateType,
    TimestampType,
    DecimalType,
    FixedType,
    ListType,
    MapType,
    StructType,
    TimeType
)
from pyiceberg.io.pyarrow import PyArrowFileIO, schema_to_pyarrow

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | bool | int | None


class StructureToIcebergConverter:
    """Class to convert JSON Structure schema to Iceberg schema."""

    def __init__(self: 'StructureToIcebergConverter'):
        self.named_type_cache: Dict[str, JsonNode] = {}
        self.id_counter = 0
        self.definitions: Dict[str, Any] = {}
        self.schema_doc: Optional[Dict[str, Any]] = None

    def get_id(self) -> int:
        """Get a unique ID for a field."""
        self.id_counter += 1
        return self.id_counter
    
    def get_fullname(self, namespace: str, name: str) -> str:
        """Get the full name of a record type."""
        return f"{namespace}.{name}" if namespace else name

    def convert_structure_to_iceberg(self, structure_schema_path: str, structure_record_type: Optional[str], output_path: str, emit_cloudevents_columns: bool=False):
        """Convert a JSON Structure schema to an Iceberg schema."""
        schema_file = structure_schema_path
        if not schema_file:
            print("Please specify the JSON Structure schema file")
            sys.exit(1)
        with open(schema_file, "r", encoding="utf-8") as f:
            schema_json = f.read()

        # Parse the schema as a JSON object
        schema = json.loads(schema_json)
        self.schema_doc = schema
        
        # Handle definitions if present
        if "definitions" in schema:
            self.definitions = schema["definitions"]
        
        # For JSON Structure, we expect an object type at the top level
        if schema.get("type") != "object":
            # Check if we have a $ref at the top level
            if "$ref" in schema:
                ref = schema["$ref"]
                schema = self.resolve_ref(ref)
            elif structure_record_type and "definitions" in schema:
                # Look for the type in definitions
                if structure_record_type in schema["definitions"]:
                    schema = schema["definitions"][structure_record_type]
                else:
                    print(f"No record type {structure_record_type} found in the JSON Structure schema definitions")
                    sys.exit(1)
            else:
                print("Expected a JSON Structure schema with type 'object' at the top level")
                sys.exit(1)

        # Get the name and properties of the top-level object
        table_name = schema.get("name", "Table")
        properties = schema.get("properties", {})
        required = schema.get("required", [])

        # Create a list to store the iceberg schema
        iceberg_fields: List[NestedField] = []

        # Append the iceberg schema with the column names and types
        for prop_name, prop_schema in properties.items():
            is_required = prop_name in required
            column_type = self.convert_structure_type_to_iceberg_type(prop_schema)
            iceberg_fields.append(
                NestedField(
                    field_id=self.get_id(),
                    name=prop_name,
                    field_type=column_type,
                    required=is_required
                ))

        if emit_cloudevents_columns:
            iceberg_fields.extend([
                NestedField(field_id=self.get_id(),
                      name="___type", field_type=StringType(), required=False),
                NestedField(field_id=self.get_id(),
                      name="___source", field_type=StringType(), required=False),
                NestedField(field_id=self.get_id(),
                      name="___id", field_type=StringType(), required=False),
                NestedField(field_id=self.get_id(),
                      name="___time", field_type=TimestampType(), required=False),
                NestedField(field_id=self.get_id(),
                      name="___subject", field_type=StringType(), required=False)
            ])

        iceberg_schema = Schema(*iceberg_fields)
        arrow_schema = schema_to_pyarrow(iceberg_schema)
        print(f"Iceberg schema created: {arrow_schema}")

        # Write to Iceberg table (for demonstration, using local file system)
        file_io = PyArrowFileIO()
        output_file = file_io.new_output("file://"+output_path)
        with output_file.create(overwrite=True) as f:
            pa.output_stream(f).write(arrow_schema.serialize().to_pybytes())

    def resolve_ref(self, ref: str) -> Dict[str, Any]:
        """Resolve a $ref reference."""
        if not ref.startswith("#/"):
            raise ValueError(f"Only local references are supported, got: {ref}")
        
        parts = ref[2:].split("/")
        current = self.schema_doc
        
        for part in parts:
            if isinstance(current, dict) and part in current:
                current = current[part]
            else:
                raise ValueError(f"Could not resolve reference: {ref}")
        
        return current

    def convert_structure_type_to_iceberg_type(self, structure_type):
        """Convert a JSON Structure type to an Iceberg type."""
        # Handle $ref
        if isinstance(structure_type, dict) and "$ref" in structure_type:
            ref = structure_type["$ref"]
            resolved = self.resolve_ref(ref)
            return self.convert_structure_type_to_iceberg_type(resolved)
        
        # Handle array of types (e.g., ["string", "null"] for nullable types)
        if isinstance(structure_type, list):
            # Filter out null from the list
            non_null_types = [t for t in structure_type if t != "null"]
            if len(non_null_types) == 1:
                # Nullable type - just use the non-null type (Iceberg handles optionality with required flag)
                return self.convert_structure_type_to_iceberg_type(non_null_types[0])
            elif len(non_null_types) > 1:
                # Union of multiple non-null types - create a struct with alternatives
                fields = []
                for i, choice in enumerate(non_null_types):
                    choice_type = self.convert_structure_type_to_iceberg_type(choice)
                    fields.append(NestedField(
                        field_id=self.get_id(),
                        name=f"option_{i}",
                        field_type=choice_type,
                        required=False
                    ))
                return StructType(*fields)
            else:
                # Only null - return string as fallback
                return StringType()
        
        # Handle dictionary with type field
        if isinstance(structure_type, dict):
            type_name = structure_type.get("type")
            
            # Handle type being an array
            if isinstance(type_name, list):
                # This is like {"type": ["string", "null"]}
                return self.convert_structure_type_to_iceberg_type(type_name)
            
            # Handle array type
            if type_name == "array":
                items = structure_type.get("items", {"type": "string"})
                return ListType(
                    element_id=self.get_id(),
                    element_type=self.convert_structure_type_to_iceberg_type(items),
                    element_required=True
                )
            
            # Handle set type (treated as array in Iceberg)
            elif type_name == "set":
                items = structure_type.get("items", {"type": "string"})
                return ListType(
                    element_id=self.get_id(),
                    element_type=self.convert_structure_type_to_iceberg_type(items),
                    element_required=True
                )
            
            # Handle map type
            elif type_name == "map":
                values = structure_type.get("values", {"type": "string"})
                return MapType(
                    key_id=self.get_id(),
                    key_type=StringType(),
                    value_id=self.get_id(),
                    value_type=self.convert_structure_type_to_iceberg_type(values),
                    value_required=True
                )
            
            # Handle tuple type (treated as struct with indexed fields)
            elif type_name == "tuple":
                items = structure_type.get("items", [])
                fields = []
                for i, item in enumerate(items):
                    fields.append(NestedField(
                        field_id=self.get_id(),
                        name=f"field_{i}",
                        field_type=self.convert_structure_type_to_iceberg_type(item),
                        required=True
                    ))
                return StructType(*fields)
            
            # Handle object type
            elif type_name == "object":
                properties = structure_type.get("properties", {})
                required = structure_type.get("required", [])
                fields = []
                
                # Handle $extends if present
                if "$extends" in structure_type:
                    extends_ref = structure_type["$extends"]
                    base_schema = self.resolve_ref(extends_ref)
                    base_properties = base_schema.get("properties", {})
                    base_required = base_schema.get("required", [])
                    
                    # Add base properties first
                    for prop_name, prop_schema in base_properties.items():
                        is_required = prop_name in base_required
                        fields.append(NestedField(
                            field_id=self.get_id(),
                            name=prop_name,
                            field_type=self.convert_structure_type_to_iceberg_type(prop_schema),
                            required=is_required
                        ))
                
                # Add own properties
                for prop_name, prop_schema in properties.items():
                    is_required = prop_name in required
                    fields.append(NestedField(
                        field_id=self.get_id(),
                        name=prop_name,
                        field_type=self.convert_structure_type_to_iceberg_type(prop_schema),
                        required=is_required
                    ))
                
                return StructType(*fields)
            
            # Handle choice type (union)
            elif type_name == "choice":
                choices = structure_type.get("choices", [])
                if isinstance(choices, list):
                    # For inline choices, create a struct with alternatives
                    fields = []
                    for i, choice in enumerate(choices):
                        choice_type = self.convert_structure_type_to_iceberg_type(choice)
                        fields.append(NestedField(
                            field_id=self.get_id(),
                            name=f"option_{i}",
                            field_type=choice_type,
                            required=False
                        ))
                    return StructType(*fields)
                elif isinstance(choices, dict):
                    # For tagged choices, create a struct with named alternatives
                    fields = []
                    for choice_name, choice_schema in choices.items():
                        choice_type = self.convert_structure_type_to_iceberg_type(choice_schema)
                        fields.append(NestedField(
                            field_id=self.get_id(),
                            name=choice_name,
                            field_type=choice_type,
                            required=False
                        ))
                    return StructType(*fields)
                else:
                    return StringType()
            
            # Handle any type
            elif type_name == "any":
                return StringType()
            
            # Handle primitive types with annotations
            elif type_name:
                return self.map_iceberg_scalar_type(type_name, structure_type)
        
        # Handle string type name directly
        elif isinstance(structure_type, str):
            return self.map_iceberg_scalar_type(structure_type, {})

        return StringType()

    def map_iceberg_scalar_type(self, type_name: str, type_schema: Dict[str, Any]):
        """Map a JSON Structure scalar type to an Iceberg scalar type."""
        # Check for decimal with precision and scale
        if type_name == "decimal":
            precision = type_schema.get("precision", 38)
            scale = type_schema.get("scale", 18)
            return DecimalType(precision, scale)
        
        # Map other primitive types
        type_mapping = {
            'null': StringType(),  # Iceberg doesn't have a null type
            'boolean': BooleanType(),
            'string': StringType(),
            'int8': IntegerType(),  # Iceberg doesn't have byte type
            'uint8': IntegerType(),
            'int16': IntegerType(),  # Iceberg doesn't have short type
            'uint16': IntegerType(),
            'int32': IntegerType(),
            'uint32': LongType(),  # Use long for unsigned int32
            'int64': LongType(),
            'uint64': LongType(),  # Iceberg doesn't distinguish signed/unsigned
            'int128': StringType(),  # No native 128-bit support
            'uint128': StringType(),
            'integer': IntegerType(),  # Generic integer
            'number': DoubleType(),  # Generic number
            'float8': FloatType(),
            'float': FloatType(),
            'float32': FloatType(),
            'binary32': FloatType(),
            'double': DoubleType(),
            'float64': DoubleType(),
            'binary64': DoubleType(),
            'decimal': DecimalType(38, 18),
            'binary': BinaryType(),
            'bytes': BinaryType(),  # Binary data
            'date': DateType(),
            'time': TimeType(),
            'datetime': TimestampType(),
            'timestamp': TimestampType(),
            'duration': LongType(),  # Store as microseconds
            'uuid': StringType(),  # Store UUID as string
            'uri': StringType(),
            'jsonpointer': StringType(),
        }
        
        return type_mapping.get(type_name, StringType())


def convert_structure_to_iceberg(structure_schema_path, structure_record_type, output_path, emit_cloudevents_columns=False):
    """Convert a JSON Structure schema to an Iceberg schema."""
    converter = StructureToIcebergConverter()
    converter.convert_structure_to_iceberg(
        structure_schema_path, structure_record_type, output_path, emit_cloudevents_columns)



# pylint: disable=line-too-long

""" StructureToJavaScript class for converting JSON Structure schema to JavaScript classes """

import json
import os
from typing import Any, Dict, List, Set, Union, Optional

from avrotize.common import pascal, process_template

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None

INDENT = '    '


def is_javascript_reserved_word(word: str) -> bool:
    """Check if word is a JavaScript reserved word"""
    reserved_words = [
        'break', 'case', 'catch', 'class', 'const', 'continue', 'debugger',
        'default', 'delete', 'do', 'else', 'export', 'extends', 'finally',
        'for', 'function', 'if', 'import', 'in', 'instanceof', 'new', 'return',
        'super', 'switch', 'this', 'throw', 'try', 'typeof', 'var', 'void',
        'while', 'with', 'yield', 'let', 'static', 'enum', 'await', 'async',
        'implements', 'interface', 'package', 'private', 'protected', 'public'
    ]
    return word in reserved_words


def is_javascript_primitive(word: str) -> bool:
    """Check if word is a JavaScript primitive"""
    primitives = ['null', 'boolean', 'number', 'string', 'Date', 'Array', 'Object']
    return word in primitives


class StructureToJavaScript:
    """Convert JSON Structure schema to JavaScript classes"""

    def __init__(self, base_package: str = '', avro_annotation=False) -> None:
        self.base_package = base_package
        self.avro_annotation = avro_annotation
        self.output_dir = os.getcwd()
        self.schema_doc: JsonNode = None
        self.generated_types: Dict[str, str] = {}
        self.generated_structure_types: Dict[str, Dict[str, Union[str, Dict, List]]] = {}
        self.type_dict: Dict[str, Dict] = {}
        self.definitions: Dict[str, Any] = {}
        self.schema_registry: Dict[str, Dict] = {}

    def get_qualified_name(self, namespace: str, name: str) -> str:
        """Concatenates namespace and name with a dot separator"""
        return f"{namespace}.{name}" if namespace != '' else name

    def concat_namespace(self, namespace: str, name: str) -> str:
        """Concatenates namespace and name with a dot separator"""
        if namespace and name:
            return f"{namespace}.{name}"
        elif namespace:
            return namespace
        else:
            return name

    def map_primitive_to_javascript(self, structure_type: str) -> str:
        """Maps JSON Structure primitive types to JavaScript types"""
        mapping = {
            'null': 'null',
            'boolean': 'boolean',
            'string': 'string',
            'integer': 'number',
            'number': 'number',
            'int8': 'number',
            'uint8': 'number',
            'int16': 'number',
            'uint16': 'number',
            'int32': 'number',
            'uint32': 'number',
            'int64': 'number',
            'uint64': 'number',
            'int128': 'bigint',
            'uint128': 'bigint',
            'float8': 'number',
            'float': 'number',
            'double': 'number',
            'binary32': 'number',
            'binary64': 'number',
            'decimal': 'string',  # JavaScript doesn't have native decimal
            'binary': 'string',   # Base64 encoded
            'date': 'Date',
            'time': 'string',     # ISO 8601 time string
            'datetime': 'Date',
            'timestamp': 'Date',
            'duration': 'string', # ISO 8601 duration string
            'uuid': 'string',
            'uri': 'string',
            'jsonpointer': 'string',
            'any': 'any'
        }
        qualified_class_name = self.get_qualified_name(self.base_package, pascal(structure_type))
        if qualified_class_name in self.generated_types:
            result = qualified_class_name
        else:
            result = mapping.get(structure_type, 'any')
        return result

    def is_javascript_primitive_type(self, js_type: str) -> bool:
        """Checks if a type is a JavaScript primitive type"""
        return js_type in ['null', 'boolean', 'number', 'string', 'Date', 'bigint', 'any']

    def safe_name(self, name: str) -> str:
        """Converts a name to a safe JavaScript name"""
        if is_javascript_reserved_word(name):
            return name + "_"
        return name

    def pascal_type_name(self, ref: str) -> str:
        """Converts a reference to a type name"""
        return pascal(ref.split('.')[-1])

    def resolve_ref(self, ref: str, context_schema: Optional[Dict] = None) -> Optional[Dict]:
        """Resolves a $ref to the actual schema definition"""
        if not ref.startswith('#/'):
            if ref in self.schema_registry:
                return self.schema_registry[ref]
            return None

        path = ref[2:].split('/')
        schema = context_schema if context_schema else self.schema_doc
        for part in path:
            if not isinstance(schema, dict) or part not in schema:
                return None
            schema = schema[part]
        return schema

    def register_schema_ids(self, schema: Dict, base_uri: str = '') -> None:
        """Recursively registers schemas with $id keywords"""
        if not isinstance(schema, dict):
            return

        if '$id' in schema:
            schema_id = schema['$id']
            if base_uri and not schema_id.startswith(('http://', 'https://', 'urn:')):
                from urllib.parse import urljoin
                schema_id = urljoin(base_uri, schema_id)
            self.schema_registry[schema_id] = schema
            base_uri = schema_id

        if 'definitions' in schema:
            for def_name, def_schema in schema['definitions'].items():
                if isinstance(def_schema, dict):
                    self.register_schema_ids(def_schema, base_uri)

        if 'properties' in schema:
            for prop_name, prop_schema in schema['properties'].items():
                if isinstance(prop_schema, dict):
                    self.register_schema_ids(prop_schema, base_uri)

        for key in ['items', 'values', 'additionalProperties']:
            if key in schema and isinstance(schema[key], dict):
                self.register_schema_ids(schema[key], base_uri)

    def convert_structure_type_to_javascript(self, class_name: str, field_name: str,
                                            structure_type: JsonNode, parent_namespace: str,
                                            import_types: Set[str]) -> str:
        """Converts JSON Structure type to JavaScript type"""
        if isinstance(structure_type, str):
            js_type = self.map_primitive_to_javascript(structure_type)
            if js_type == structure_type and not self.is_javascript_primitive_type(js_type):
                import_types.add(js_type)
                return self.pascal_type_name(js_type)
            return js_type
        elif isinstance(structure_type, list):
            # Handle type unions
            non_null_types = [t for t in structure_type if t != 'null']
            if len(non_null_types) == 1:
                inner_type = self.convert_structure_type_to_javascript(
                    class_name, field_name, non_null_types[0], parent_namespace, import_types)
                if 'null' in structure_type:
                    return f'{inner_type}|null'
                return inner_type
            else:
                union_types = [self.convert_structure_type_to_javascript(
                    class_name, field_name, t, parent_namespace, import_types) for t in non_null_types]
                return '|'.join(union_types)
        elif isinstance(structure_type, dict):
            # Handle $ref
            if '$ref' in structure_type:
                ref_schema = self.resolve_ref(structure_type['$ref'], self.schema_doc)
                if ref_schema:
                    ref_path = structure_type['$ref'].split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    ref = self.generate_class_or_choice(ref_schema, ref_namespace, write_file=True, explicit_name=type_name)
                    import_types.add(ref)
                    return ref.split('.')[-1]
                return 'any'

            # Handle enum keyword
            if 'enum' in structure_type:
                enum_ref = self.generate_enum(structure_type, field_name, parent_namespace, write_file=True)
                import_types.add(enum_ref)
                return enum_ref.split('.')[-1]

            # Handle type keyword
            if 'type' not in structure_type:
                return 'any'

            struct_type = structure_type['type']

            # Handle complex types
            if struct_type == 'object':
                class_ref = self.generate_class(structure_type, parent_namespace, write_file=True)
                import_types.add(class_ref)
                return class_ref.split('.')[-1]
            elif struct_type == 'array':
                items_type = self.convert_structure_type_to_javascript(
                    class_name, field_name+'Item', structure_type.get('items', {'type': 'any'}),
                    parent_namespace, import_types)
                return f'Array<{items_type}>'
            elif struct_type == 'set':
                items_type = self.convert_structure_type_to_javascript(
                    class_name, field_name+'Item', structure_type.get('items', {'type': 'any'}),
                    parent_namespace, import_types)
                return f'Set<{items_type}>'
            elif struct_type == 'map':
                values_type = self.convert_structure_type_to_javascript(
                    class_name, field_name+'Value', structure_type.get('values', {'type': 'any'}),
                    parent_namespace, import_types)
                return f'Object<string, {values_type}>'
            elif struct_type == 'choice':
                return self.generate_choice(structure_type, parent_namespace, write_file=True, import_types=import_types)
            elif struct_type == 'tuple':
                return self.generate_tuple(structure_type, parent_namespace, write_file=True, import_types=import_types)
            else:
                return self.convert_structure_type_to_javascript(class_name, field_name, struct_type, parent_namespace, import_types)
        return 'any'

    def generate_class_or_choice(self, structure_schema: Dict, parent_namespace: str,
                                 write_file: bool = True, explicit_name: str = '') -> str:
        """Generates a Class or Choice"""
        struct_type = structure_schema.get('type', 'object')
        if struct_type == 'object':
            return self.generate_class(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'choice':
            return self.generate_choice(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'tuple':
            return self.generate_tuple(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        return 'any'

    def generate_class(self, structure_schema: Dict, parent_namespace: str,
                      write_file: bool, explicit_name: str = '') -> str:
        """Generates a JavaScript class from JSON Structure object type"""
        import_types: Set[str] = set()

        # Get name and namespace
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedClass'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_package, schema_namespace)
        qualified_name = self.get_qualified_name(namespace, class_name)

        if qualified_name in self.generated_types:
            return qualified_name

        # Check if this is an abstract type
        is_abstract = structure_schema.get('abstract', False)

        # Handle inheritance ($extends)
        base_class = None
        base_class_name = None
        if '$extends' in structure_schema:
            base_ref = structure_schema['$extends']
            if isinstance(self.schema_doc, dict):
                base_schema = self.resolve_ref(base_ref, self.schema_doc)
                if base_schema:
                    ref_path = base_ref.split('/')
                    base_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    base_class = self.generate_class(base_schema, ref_namespace, write_file=True, explicit_name=base_name)
                    import_types.add(base_class)
                    base_class_name = pascal(base_name)

        # Generate properties
        properties = structure_schema.get('properties', {})
        required_props = structure_schema.get('required', [])

        # Collect fields for template
        fields = []
        static_fields = []

        for prop_name, prop_schema in properties.items():
            field_name = self.safe_name(prop_name)
            field_doc = prop_schema.get('description', prop_schema.get('doc', ''))

            # Check if this is a const field
            if 'const' in prop_schema:
                # Const fields are static properties
                const_value = self.format_const_value(prop_schema['const'])
                static_fields.append({
                    'name': field_name,
                    'value': const_value,
                    'docstring': field_doc
                })
                continue

            # Determine if required
            is_required = prop_name in required_props if not isinstance(required_props, list) or \
                         len(required_props) == 0 or not isinstance(required_props[0], list) else \
                         any(prop_name in req_set for req_set in required_props)

            # Get property type
            prop_type = self.convert_structure_type_to_javascript(
                class_name, field_name, prop_schema, schema_namespace, import_types)

            # Add default value if present
            if 'default' in prop_schema:
                default_val = self.format_default_value(prop_schema['default'], prop_type)
            elif not is_required:
                default_val = 'null'
            else:
                default_val = 'undefined'

            fields.append({
                'name': field_name,
                'type': prop_type,
                'default_value': default_val,
                'docstring': field_doc
            })

        # Get docstring
        doc = structure_schema.get('description', structure_schema.get('doc', class_name))

        # Generate imports dictionary
        imports = {}
        for import_type in import_types:
            import_type_package = import_type.rsplit('.', 1)[0]
            import_type_type = pascal(import_type.split('.')[-1])
            import_type_package = import_type_package.replace('.', '/')
            namespace_path = namespace.replace('.', '/')

            if import_type_package:
                import_type_package = os.path.relpath(import_type_package, namespace_path).replace(os.sep, '/')
                if not import_type_package.startswith('.'):
                    import_type_package = f'./{import_type_package}'
                imports[import_type_type] = f'{import_type_package}/{import_type_type}'
            else:
                imports[import_type_type] = f'./{import_type_type}'

        # Generate class definition using template
        class_definition = process_template(
            "structuretojs/class_core.js.jinja",
            class_name=class_name,
            docstring=doc,
            fields=fields,
            static_fields=static_fields,
            imports=imports,
            is_abstract=is_abstract,
            base_class_name=base_class_name
        )

        if write_file:
            self.write_to_file(namespace, class_name, class_definition)
            # Generate test file for this class
            self.generate_test_class(namespace, class_name, fields, import_types, prop_schema=structure_schema)

        self.generated_types[qualified_name] = 'class'
        self.generated_structure_types[qualified_name] = structure_schema
        return qualified_name

    def generate_enum(self, structure_schema: Dict, field_name: str, parent_namespace: str,
                     write_file: bool) -> str:
        """Generates a JavaScript enum from JSON Structure enum"""
        class_name = pascal(structure_schema.get('name', field_name + 'Enum'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_package, schema_namespace)
        qualified_name = self.get_qualified_name(namespace, class_name)

        if qualified_name in self.generated_types:
            return qualified_name

        enum_values = structure_schema.get('enum', [])
        symbols = [str(symbol) if not is_javascript_reserved_word(str(symbol)) else str(symbol) + "_"
                   for symbol in enum_values]
        symbol_values = [str(val) for val in enum_values]

        doc = structure_schema.get('description', structure_schema.get('doc', f'A {class_name} enum.'))

        enum_definition = process_template(
            "structuretojs/enum_core.js.jinja",
            class_name=class_name,
            docstring=doc,
            symbols=symbols,
            symbol_values=symbol_values,
        )

        if write_file:
            self.write_to_file(namespace, class_name, enum_definition)
            # Generate test file for this enum
            self.generate_test_enum(namespace, class_name, enum_values)

        self.generated_types[qualified_name] = 'enum'
        self.generated_structure_types[qualified_name] = structure_schema
        return qualified_name

    def generate_choice(self, structure_schema: Dict, parent_namespace: str,
                       write_file: bool, explicit_name: str = '', import_types: Optional[Set[str]] = None) -> str:
        """Generates a JavaScript Union type from JSON Structure choice"""
        if import_types is None:
            import_types = set()

        # Generate types for each choice
        choice_types = []
        choices = structure_schema.get('choices', {})

        for choice_key, choice_schema in choices.items():
            if isinstance(choice_schema, dict):
                if '$ref' in choice_schema:
                    ref_schema = self.resolve_ref(choice_schema['$ref'], self.schema_doc)
                    if ref_schema:
                        ref_path = choice_schema['$ref'].split('/')
                        ref_name = ref_path[-1]
                        ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                        qualified_name = self.generate_class(ref_schema, ref_namespace, write_file=True, explicit_name=ref_name)
                        import_types.add(qualified_name)
                        choice_types.append(qualified_name.split('.')[-1])
                elif 'type' in choice_schema:
                    js_type = self.convert_structure_type_to_javascript(
                        explicit_name, choice_key, choice_schema, parent_namespace, import_types)
                    choice_types.append(js_type)

        # Return union type string
        if len(choice_types) == 0:
            return 'any'
        elif len(choice_types) == 1:
            return choice_types[0]
        else:
            return '|'.join(choice_types)

    def generate_tuple(self, structure_schema: Dict, parent_namespace: str,
                      write_file: bool, explicit_name: str = '', import_types: Optional[Set[str]] = None) -> str:
        """Generates a JavaScript Array type for tuples"""
        if import_types is None:
            import_types = set()

        # Tuples in JavaScript are represented as arrays
        # For now, return Array<any> as tuples need special handling
        return 'Array<any>'

    def format_const_value(self, value: Any) -> str:
        """Formats a const value for JavaScript"""
        if value is None:
            return 'null'
        elif isinstance(value, bool):
            return 'true' if value else 'false'
        elif isinstance(value, str):
            return f'"{value}"'
        elif isinstance(value, (int, float)):
            return str(value)
        elif isinstance(value, list):
            items = ', '.join([self.format_const_value(v) for v in value])
            return f'[{items}]'
        elif isinstance(value, dict):
            items = ', '.join([f'"{k}": {self.format_const_value(v)}' for k, v in value.items()])
            return f'{{{items}}}'
        return 'null'

    def format_default_value(self, value: Any, js_type: str) -> str:
        """Formats a default value for JavaScript"""
        return self.format_const_value(value)

    def write_to_file(self, namespace: str, name: str, content: str):
        """Writes JavaScript class to file"""
        directory_path = os.path.join(self.output_dir, namespace.replace('.', os.sep), 'src')
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)

        file_path = os.path.join(directory_path, f"{name}.js")
        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(content)

    def write_test_to_file(self, namespace: str, name: str, content: str):
        """Writes test content to a file in the test directory"""
        directory_path = os.path.join(self.output_dir, namespace.replace('.', os.sep), 'test')
        if not os.path.exists(directory_path):
            os.makedirs(directory_path, exist_ok=True)

        file_path = os.path.join(directory_path, f'test_{name}.js')
        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(content)

    def generate_test_value(self, field_type: str, field_name: str) -> str:
        """Generate appropriate test value based on field type"""
        if field_type == 'string':
            return f'"test_{field_name}"'
        elif field_type in ['number', 'int', 'integer', 'float', 'double']:
            return '42'
        elif field_type == 'boolean':
            return 'true'
        elif field_type == 'Date':
            return 'new Date("2024-01-01")'
        elif field_type == 'null':
            return 'null'
        elif field_type.startswith('Array'):
            return '[]'
        elif field_type.startswith('Set'):
            return 'new Set()'
        elif field_type.startswith('Map') or field_type == 'Object':
            return '{}'
        else:
            return f'new {field_type}()'

    def generate_test_class(self, namespace: str, class_name: str, fields: List[Dict],
                           import_types: Set[str], prop_schema: Dict):
        """Generate test file for a class"""
        test_imports = {}
        for import_type in import_types:
            import_type_type = pascal(import_type.split('.')[-1])
            import_type_package = import_type.rsplit('.', 1)[0].replace('.', '/')
            namespace_path = namespace.replace('.', '/')

            if import_type_package:
                rel_path = os.path.relpath(import_type_package, namespace_path).replace(os.sep, '/')
                if not rel_path.startswith('.'):
                    rel_path = f'./{rel_path}'
                test_imports[import_type_type] = f'../{rel_path}/{import_type_type}'
            else:
                test_imports[import_type_type] = f'./{import_type_type}'

        test_fields = []
        for field in fields:
            test_value = self.generate_test_value(field['type'], field['name'])
            test_fields.append({
                'name': field['name'],
                'type': field['type'],
                'test_value': test_value,
                'docstring': field.get('docstring', '')
            })

        # Class path should just be the class name since src/test structure mirrors each other
        class_path = class_name

        test_definition = process_template(
            "structuretojs/test_class.js.jinja",
            class_name=class_name,
            class_path=class_path,
            fields=test_fields,
            test_imports=test_imports
        )

        self.write_test_to_file(namespace, class_name, test_definition)

    def generate_test_enum(self, namespace: str, enum_name: str, enum_values: List):
        """Generate test file for an enum"""
        expected_values = ', '.join([f'"{val}"' for val in enum_values])

        # Enum path should just be the enum name
        enum_path = enum_name

        test_definition = process_template(
            "structuretojs/test_enum.js.jinja",
            enum_name=enum_name,
            enum_path=enum_path,
            expected_values=expected_values
        )

        self.write_test_to_file(namespace, enum_name, test_definition)

    def process_definitions(self, definitions: Dict, namespace_path: str) -> None:
        """Processes the definitions section recursively"""
        for name, definition in definitions.items():
            if isinstance(definition, dict):
                if 'type' in definition:
                    current_namespace = self.concat_namespace(namespace_path, '')
                    check_namespace = self.concat_namespace(self.base_package, current_namespace)
                    check_name = pascal(name)
                    check_ref = self.get_qualified_name(check_namespace, check_name)
                    if check_ref not in self.generated_types:
                        self.generate_class_or_choice(definition, current_namespace, write_file=True, explicit_name=name)
                else:
                    new_namespace = self.concat_namespace(namespace_path, name)
                    self.process_definitions(definition, new_namespace)

    def convert_schemas(self, structure_schemas: List, output_dir: str):
        """Converts JSON Structure schemas to JavaScript classes"""
        self.output_dir = output_dir
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir, exist_ok=True)

        # Register all schema IDs first
        for structure_schema in structure_schemas:
            if isinstance(structure_schema, dict):
                self.register_schema_ids(structure_schema)

        for structure_schema in structure_schemas:
            if not isinstance(structure_schema, dict):
                continue

            self.schema_doc = structure_schema

            if 'definitions' in structure_schema:
                self.definitions = structure_schema['definitions']

            # Process root type first
            if 'type' in structure_schema:
                self.generate_class_or_choice(structure_schema, '', write_file=True)
            elif '$root' in structure_schema:
                root_ref = structure_schema['$root']
                root_schema = self.resolve_ref(root_ref, structure_schema)
                if root_schema:
                    ref_path = root_ref.split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else ''
                    self.generate_class_or_choice(root_schema, ref_namespace, write_file=True, explicit_name=type_name)

            # Process definitions
            if 'definitions' in structure_schema:
                self.process_definitions(self.definitions, '')

        # Generate test runner after all classes and enums are generated
        self.generate_test_runner()

    def generate_test_runner(self):
        """Generate the test runner script"""
        test_runner = process_template(
            "structuretojs/test_runner.js.jinja"
        )

        # Write test runner to root test directory
        test_dir = os.path.join(self.output_dir, self.base_package.replace('.', os.sep), 'test')
        if not os.path.exists(test_dir):
            os.makedirs(test_dir, exist_ok=True)

        test_runner_path = os.path.join(test_dir, 'test_runner.js')
        with open(test_runner_path, 'w', encoding='utf-8') as file:
            file.write(test_runner)

    def convert(self, structure_schema_path: str, output_dir: str):
        """Converts JSON Structure schema to JavaScript classes"""
        with open(structure_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        if isinstance(schema, dict):
            schema = [schema]
        return self.convert_schemas(schema, output_dir)


def convert_structure_to_javascript(structure_schema_path, js_file_path, package_name='', avro_annotation=False):
    """Converts JSON Structure schema to JavaScript classes"""
    if not package_name:
        package_name = os.path.splitext(os.path.basename(structure_schema_path))[0].lower().replace('-', '_')

    structure_to_js = StructureToJavaScript(package_name, avro_annotation=avro_annotation)
    structure_to_js.convert(structure_schema_path, js_file_path)


def convert_structure_schema_to_javascript(structure_schema, js_file_path, package_name='', avro_annotation=False):
    """Converts JSON Structure schema to JavaScript classes"""
    structure_to_js = StructureToJavaScript(package_name, avro_annotation=avro_annotation)
    if isinstance(structure_schema, dict):
        structure_schema = [structure_schema]
    structure_to_js.convert_schemas(structure_schema, js_file_path)
